;; gorilla-repl.fileformat = 1

;; **
;;; # Chaos
;; **

;; @@
(ns worksheets.chaos
    (:require [gorilla-plot.core :as plot]
           [anglican.core :refer [doquery]]
           [bopp.core :refer :all]
           [bopp.helper-functions :refer [argmax]]
           [clojure-csv.core :refer :all]
           [clojure.data.csv :as csv]
           [clojure.java.io :as io]
           [clojure.core.matrix :as m
            :refer
            [matrix identity-matrix zero-vector
             shape inverse transpose
             mul mmul add sub div]]
           [clojure.data.json :as json])
  (:use
    clojure.repl
        [anglican
          runtime
          emit
          smc
          stat
          [state :only [get-predicts get-log-weight]]
          [inference :only [log-marginal rand-roulette]]]))
;; @@
;; =>
;;; {"type":"html","content":"<span class='clj-nil'>nil</span>","value":"nil"}
;; <=

;; **
;;; ## The Problem
;;; 
;;; As an example application we consider the case of optimizing the transition function parameters of an extended Kalman filter for the tracking of a chaotic attractor.  Chaotic attractors present an interesting case for tracking problems as, although their underlying dynamics are strictly deterministic with bounded trajectories, neighbouring trajectories diverge exponentially. Therefore regardless of the available precision, a trajectory cannot be indefinitely extrapolated to within a given accuracy and probabilistic methods such as the extended Kalman filter must be incorporated. From an empirical perspective, this forms a challenging optimization problem as the target transpires to be multi-modal, has variations at different length scales and has local minima close to the global maximum.
;; **

;; **
;;; Suppose we observe a noisy signal @@y\_t \in \mathbb{R}^{K}, \; t = 1,2,\dots@@ in some @@K@@ dimensional observation space which we believe has a lower dimensional latent space @@x\_t \in \mathbb{R}^{D}@@ corresponding to a chaotic attractor of known type but with unknown parameters.  Given observations up to some time @@T@@, we wish to performance inference over the latent space using an extended Kalman filter as defined by
;;; 
;;; @@\begin{align}
;;; x\_0 \sim & \mathcal{N} \left(\mu\_0, \sigma\_0 I\right) \\\\
;;; x\_t = & A \left(x\_{t-1}, \theta\right)+\delta\_{t-1}, \quad \delta\_{t-1} \sim \mathcal{N} \left(0, \sigma\_q I\right) \\\\
;;; y\_t = & C x\_{t}+\varepsilon\_{t}, \quad \varepsilon\_{t} \sim \mathcal{N} \left(0, \sigma\_y I\right)
;;; \end{align}@@
;;; 
;;; where @@I@@ is the identity matrix, @@C@@ is a known @@K \times D@@ matrix,  @@\mu\_0@@ is the expected starting position, and @@\sigma\_0, \sigma\_q@@ and @@\sigma\_y@@ are all scalars which are assumed to be known.  The transition function @@A \left(\cdot,\cdot\right)@@ is
;;; 
;;; @@\begin{align}
;;; 	x\_{t,1} = & \; \sin \left(\beta x\_{t-1,2}\right)-\cos\left(\frac{5x\_{t-1,1}}{2}\right)x\_{t-1,3}  \\\\
;;;     	x\_{t,2} = & \;-\sin \left(\frac{3x\_{t-1,1}}{2}\right)x\_{t-1,3}-\cos\left(\eta x\_{t-1,2}\right) \\\\
;;; x\_{t,3} = & \; \sin \left(x\_{t-1,1}\right)
;;; \end{align}@@
;;; 	
;;; 
;;; corresponding to a type of Pickover attractor with unknown parameters @@\theta = \\{\beta,\eta\\}@@ which we wish to optimize.  Note that @@\eta@@ and @@-\eta@@ will give the same behaviour. Lets start by defining some helper functions, defining the variables, and importing the data.                         
;; **

;; @@
(defn get-data [file n-points]
  (->> (str file)
       io/resource
       io/reader
       slurp
       csv/read-csv
       (into [])
       (mapv #(mapv read-string %))
       (take n-points)
       (into [])))

(def T 50) ;; Less that paper for speed purposes
(def observations (into [] (get-data "data/chaos/y1.csv" T)))

(defn A
  [beta nu x y z]
  "Chaotic transition according to the pickover attractor"
  [(- (sin (* beta y)) (* z (cos (* 5/2 x))))
   (- (* z (sin (* -3/2 x))) (cos (* nu y)))
   (sin x)])

;; Define the other parameters

(def C
  (transpose (matrix
   [[0.0243087960714491	0.0168113871672383	0.0691772445397626	1.57387136968515e-05	0.303732278352682	0.0618563110870659	4.95623561031004e-06	1.89039711478951e-08	0.00113995322996380	0.00515810614718469	0.00341658550834822	7.85881265878172e-07	1.30076659987469e-16	0.291018005156189	3.67275441629293e-10	3.20483237445295e-05	1.63410289930715e-18	1.65003111575406e-05	0.211775125663058	0.0115361583403371]
    [2.03156858097514e-14	1.61789389343840e-08	0.0518864591774790	3.77174213741851e-10	0.000580831681939922	7.97405604743620e-30	0.0660259803416299	6.29137955059571e-12	0.000500238717434247	2.78157203235329e-05	3.67744070950491e-05	0.706675337480128	0.00104581188589214	7.94060469213274e-06	0.0634211844981511	7.77022364283257e-19	0.0558005175443096	0.0539415131008550	3.43420457005632e-05	1.52362319447776e-05]
    [2.02934951680463e-05	0.0819514426812765	0.00587746530272417	1.12942910915840e-08	7.07682025375137e-07	0.000219700599512478	0.229511658185978	0.209524004136316	2.35546751056699e-19	0.286199014662333	0.00126960750520102	0.0892967894625532	0.00508817870175021	2.91561618504951e-24	0.0895862772514099	6.47343828059709e-06	5.67384242548235e-09	0.000224094066002085	0.00122301803539802	1.25782593868869e-06]])))

(def mu0 [1.5 -0.8 -0.2]) ;; Different to paper to allow fewer particles and thus faster example
(def sig-0 0.1)
(def sig-q 0.01)
(def sig-y 0.2)

;; For speed, define a custom observation distribution
(defdist obs-dist
  [x] ; distribution parameters
  [dist (normal 0 sig-y)]        ; auxiliary bindings
  (sample* [this] (mapv #(+ % (sample* dist)) (mmul C x)))
  (observe* [this value]
           (reduce + (map #(observe* dist (- %1 %2))
                          (mmul C x)
                          value))))
;; @@
;; =>
;;; {"type":"html","content":"<span class='clj-unkown'>#multifn[print-method 0x480e660d]</span>","value":"#multifn[print-method 0x480e660d]"}
;; <=

;; **
;;; ## Solution using BOPP
;;; 
;;; We can solve this problem using BOPP by coding the problem and running automatic MMAP estimation to optimize @@\theta@@.
;; **

;; @@
(with-primitive-procedures [matrix add obs-dist A]
   (defopt kalman-chaos-opt
    [observations] ;; Fixed inputs
    [beta nu] ;; Parameters to optimize
    (let [beta (sample (uniform-continuous -3 3)) ;; These effectly represent the bounds on our optimization
          nu (sample (uniform-continuous 0 3))
          delta-dist (normal 0 sig-q)
          trans-sample (fn [x] (into [] (map #(+ % (sample delta-dist)) ;; mapv not cps transformed by default
                       		  		 		  (A beta nu (first x) (second x) (nth x 2)))))
          initial-vals (add mu0 (into [] (repeatedly 3 #(sample (normal 0 sig-y)))))]
      (reduce (fn [states obs]
                (let [prev-state (peek states);; sample next state
                      state (if prev-state
                                (trans-sample prev-state)
                                initial-vals)]
                  (observe (count states) (obs-dist state) obs) ;; observe next data point (when available)
                  (conj states state))) ;; append state to sequence and continue with next obs       
         [] ;; start with empty sequence         
         observations)))) ;; loop over data, return states
;; @@
;; =>
;;; {"type":"html","content":"<span class='clj-var'>#&#x27;worksheets.chaos/kalman-chaos-opt</span>","value":"#'worksheets.chaos/kalman-chaos-opt"}
;; <=

;; **
;;; Next we can call BOPP directly on this to optimize beta and nu.  Warning this takes some time, MMAP is hard!  Results are likely to be a bit worse that in the paper as we have speed-option set to :fast instead of :careful here.
;; **

;; @@
(def samples (->> (doopt :smc 
                         kalman-chaos-opt
                         [observations]
                         250 ;; Number of particles
                         :speed-option :fast
                         :bo-options {:verbose 1})
                (take 50) ;; Number of optimization iterations to do
                doall))
;; @@

;; @@
(def theta-cum-best (mapv first samples))
(def mu-log-ests (mapv second samples))
(def example-trajec (mapv last samples))
(def inferred-trajetories (map first (map #(nth % 2) samples)))
;; @@
;; =>
;;; {"type":"html","content":"<span class='clj-var'>#&#x27;worksheets.chaos/inferred-trajetories</span>","value":"#'worksheets.chaos/inferred-trajetories"}
;; <=

;; **
;;; Log marginal as a function of BOPP iteration.  Note this can go down because BOPP is trying to estimate the variance in the estimator at the same time and so it may decide that a previously supposedly good result was more chance then good parameters, i.e. the GP concludes that there is high noise and that the expected maximum is significantly below the highest evaluated value.
;; **

;; @@
(plot/list-plot (map vector (range 1 (inc (count samples))) (map second samples))
                :x-title "Iteration"
                :y-title "Log marginal p(y)"
                :joined true)
;; @@
;; =>
;;; {"type":"vega","content":{"width":400,"height":247.2187957763672,"padding":{"top":10,"left":55,"bottom":40,"right":10},"data":[{"name":"5507688b-5455-4eba-a3c8-a23067224177","values":[{"x":1,"y":313.5185568335571},{"x":2,"y":399.96538468913843},{"x":3,"y":399.95680860775786},{"x":4,"y":399.94846472934296},{"x":5,"y":399.94089960458683},{"x":6,"y":467.1189259438181},{"x":7,"y":467.0479223364398},{"x":8,"y":467.11661943325555},{"x":9,"y":467.0249326963825},{"x":10,"y":471.4744656794019},{"x":11,"y":471.7102199110657},{"x":12,"y":470.60995988405955},{"x":13,"y":470.996542537961},{"x":14,"y":470.70599967453745},{"x":15,"y":470.9411815549188},{"x":16,"y":471.85226564589493},{"x":17,"y":472.28722720973815},{"x":18,"y":480.77536184885264},{"x":19,"y":585.524468089941},{"x":20,"y":618.3710424998832},{"x":21,"y":619.6210656596608},{"x":22,"y":602.4947162396337},{"x":23,"y":576.1650551221047},{"x":24,"y":619.302858387333},{"x":25,"y":595.5710128244214},{"x":26,"y":619.4327557120702},{"x":27,"y":600.8809492097532},{"x":28,"y":619.6483782527667},{"x":29,"y":591.8588192096956},{"x":30,"y":608.9798462056505},{"x":31,"y":549.3426860286805},{"x":32,"y":601.3880312253932},{"x":33,"y":617.0081874785553},{"x":34,"y":589.3173093041021},{"x":35,"y":591.7850391388502},{"x":36,"y":577.5680128385472},{"x":37,"y":588.1832744916405},{"x":38,"y":599.0452767353243},{"x":39,"y":580.8211346809599},{"x":40,"y":567.5001112196575},{"x":41,"y":552.9284330163281},{"x":42,"y":594.4515134138096},{"x":43,"y":592.8091920524978},{"x":44,"y":628.6595675891303},{"x":45,"y":650.1528401096248},{"x":46,"y":658.5638467846649},{"x":47,"y":648.9090819068475},{"x":48,"y":644.2064036322622},{"x":49,"y":643.5054312521519},{"x":50,"y":647.5976070727606}]}],"marks":[{"type":"line","from":{"data":"5507688b-5455-4eba-a3c8-a23067224177"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"#FF29D2"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}}],"scales":[{"name":"x","type":"linear","range":"width","zero":false,"domain":{"data":"5507688b-5455-4eba-a3c8-a23067224177","field":"data.x"}},{"name":"y","type":"linear","range":"height","nice":true,"zero":false,"domain":{"data":"5507688b-5455-4eba-a3c8-a23067224177","field":"data.y"}}],"axes":[{"type":"x","scale":"x","title":"Iteration","titleOffset":30},{"type":"y","scale":"y","title":"Log marginal p(y)","titleOffset":45}]},"value":"#gorilla_repl.vega.VegaView{:content {:width 400, :height 247.2188, :padding {:top 10, :left 55, :bottom 40, :right 10}, :data [{:name \"5507688b-5455-4eba-a3c8-a23067224177\", :values ({:x 1, :y 313.5185568335571} {:x 2, :y 399.96538468913843} {:x 3, :y 399.95680860775786} {:x 4, :y 399.94846472934296} {:x 5, :y 399.94089960458683} {:x 6, :y 467.1189259438181} {:x 7, :y 467.0479223364398} {:x 8, :y 467.11661943325555} {:x 9, :y 467.0249326963825} {:x 10, :y 471.4744656794019} {:x 11, :y 471.7102199110657} {:x 12, :y 470.60995988405955} {:x 13, :y 470.996542537961} {:x 14, :y 470.70599967453745} {:x 15, :y 470.9411815549188} {:x 16, :y 471.85226564589493} {:x 17, :y 472.28722720973815} {:x 18, :y 480.77536184885264} {:x 19, :y 585.524468089941} {:x 20, :y 618.3710424998832} {:x 21, :y 619.6210656596608} {:x 22, :y 602.4947162396337} {:x 23, :y 576.1650551221047} {:x 24, :y 619.302858387333} {:x 25, :y 595.5710128244214} {:x 26, :y 619.4327557120702} {:x 27, :y 600.8809492097532} {:x 28, :y 619.6483782527667} {:x 29, :y 591.8588192096956} {:x 30, :y 608.9798462056505} {:x 31, :y 549.3426860286805} {:x 32, :y 601.3880312253932} {:x 33, :y 617.0081874785553} {:x 34, :y 589.3173093041021} {:x 35, :y 591.7850391388502} {:x 36, :y 577.5680128385472} {:x 37, :y 588.1832744916405} {:x 38, :y 599.0452767353243} {:x 39, :y 580.8211346809599} {:x 40, :y 567.5001112196575} {:x 41, :y 552.9284330163281} {:x 42, :y 594.4515134138096} {:x 43, :y 592.8091920524978} {:x 44, :y 628.6595675891303} {:x 45, :y 650.1528401096248} {:x 46, :y 658.5638467846649} {:x 47, :y 648.9090819068475} {:x 48, :y 644.2064036322622} {:x 49, :y 643.5054312521519} {:x 50, :y 647.5976070727606})}], :marks [{:type \"line\", :from {:data \"5507688b-5455-4eba-a3c8-a23067224177\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"#FF29D2\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}}], :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"5507688b-5455-4eba-a3c8-a23067224177\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"5507688b-5455-4eba-a3c8-a23067224177\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Iteration\", :titleOffset 30} {:type \"y\", :scale \"y\", :title \"Log marginal p(y)\", :titleOffset 45}]}}"}
;; <=

;; **
;;; Ground truth (blue) and inferred (red) trajectories. Left to right: dimensions 1 to 3. After **1 iteration of BOPP**.
;; **

;; @@
(def ground-truth-trajectory (transpose (into [] (get-data "data/chaos/x1.csv" T))))
(def inferred-trajectories (map (fn [x] (transpose (first x))) (map #(nth % 2) samples)))
;; @@
;; =>
;;; {"type":"html","content":"<span class='clj-var'>#&#x27;worksheets.chaos/inferred-trajectories</span>","value":"#'worksheets.chaos/inferred-trajectories"}
;; <=

;; @@
(def plot-colors ["blue" "red" "#05A" "#A05" "#0A5" "#50A" "#A50" "#5A0"])
(mapv 
  #(plot/compose
     (plot/list-plot %1
                     :joined true 
                     :color (first plot-colors)
                     :plot-size 250
                     :x-title "Time step")
     (plot/list-plot %2
                     :joined true 
                     :color (second plot-colors)))
  ground-truth-trajectory
  (first inferred-trajectories))
;; @@
;; =>
;;; {"type":"list-like","open":"<span class='clj-vector'>[</span>","close":"<span class='clj-vector'>]</span>","separator":" ","items":[{"type":"vega","content":{"width":250,"height":154.51174926757812,"padding":{"top":10,"left":55,"bottom":40,"right":10},"scales":[{"name":"x","type":"linear","range":"width","zero":false,"domain":{"data":"0585429e-c9cb-4cc4-8a66-dacca3e9363a","field":"data.x"}},{"name":"y","type":"linear","range":"height","nice":true,"zero":false,"domain":{"data":"0585429e-c9cb-4cc4-8a66-dacca3e9363a","field":"data.y"}}],"axes":[{"type":"x","scale":"x","title":"Time step","titleOffset":30},{"type":"y","scale":"y"}],"data":[{"name":"0585429e-c9cb-4cc4-8a66-dacca3e9363a","values":[{"x":0,"y":1.4177},{"x":1,"y":0.76228},{"x":2,"y":1.079},{"x":3,"y":-0.20891},{"x":4,"y":-0.6084},{"x":5,"y":1.0057},{"x":6,"y":0.51189},{"x":7,"y":-0.26634},{"x":8,"y":-0.8711},{"x":9,"y":-1.128},{"x":10,"y":0.030323},{"x":11,"y":1.5069},{"x":12,"y":0.41506},{"x":13,"y":0.23892},{"x":14,"y":0.54717},{"x":15,"y":0.92183},{"x":16,"y":1.1968},{"x":17,"y":1.6507},{"x":18,"y":0.90201},{"x":19,"y":1.6279},{"x":20,"y":-0.15032},{"x":21,"y":-0.87007},{"x":22,"y":0.89195},{"x":23,"y":0.53004},{"x":24,"y":-0.44385},{"x":25,"y":-0.62716},{"x":26,"y":-0.99799},{"x":27,"y":0.24588},{"x":28,"y":1.5889},{"x":29,"y":0.51156},{"x":30,"y":0.19102},{"x":31,"y":0.50977},{"x":32,"y":0.9414},{"x":33,"y":1.2945},{"x":34,"y":1.5499},{"x":35,"y":1.3471},{"x":36,"y":1.8116},{"x":37,"y":0.34641},{"x":38,"y":0.10246},{"x":39,"y":-0.37094},{"x":40,"y":-0.30507},{"x":41,"y":1.0961},{"x":42,"y":0.67648},{"x":43,"y":0.99356},{"x":44,"y":0.012639},{"x":45,"y":-0.3428},{"x":46,"y":0.77221},{"x":47,"y":0.58104},{"x":48,"y":0.9016},{"x":49,"y":0.58413}]},{"name":"dc1ebe84-c3ed-41f5-9df8-e2831f15109f","values":[{"x":0,"y":1.3387421635344436},{"x":1,"y":0.47872890986650246},{"x":2,"y":-0.34134570045021984},{"x":3,"y":-0.16034407351471264},{"x":4,"y":0.23278263418044728},{"x":5,"y":0.08937256267855709},{"x":6,"y":-0.19924327655020221},{"x":7,"y":-0.012050489565514359},{"x":8,"y":0.1440881803645548},{"x":9,"y":0.054289186459479735},{"x":10,"y":-0.09192097157915488},{"x":11,"y":0.00705905140184483},{"x":12,"y":0.12375150517624058},{"x":13,"y":0.02440629257132847},{"x":14,"y":-0.09111963326647114},{"x":15,"y":0.04548400618581666},{"x":16,"y":0.08906195165088954},{"x":17,"y":0.05475261757205917},{"x":18,"y":-0.13442502250557192},{"x":19,"y":-0.05066028402771136},{"x":20,"y":0.2092552812771087},{"x":21,"y":-0.017591681548372852},{"x":22,"y":-0.2392730147654148},{"x":23,"y":0.08860668830440048},{"x":24,"y":0.2005317917942492},{"x":25,"y":-0.004314958816411618},{"x":26,"y":-0.23042520244617012},{"x":27,"y":0.07060670823721525},{"x":28,"y":0.18565453792717046},{"x":29,"y":-0.03559618378901941},{"x":30,"y":-0.1740038594564115},{"x":31,"y":0.07817094628555893},{"x":32,"y":0.20731280618983428},{"x":33,"y":-0.09500327490783189},{"x":34,"y":-0.1366131547318104},{"x":35,"y":0.06309104086870862},{"x":36,"y":0.21343708472650297},{"x":37,"y":-0.06358634686274946},{"x":38,"y":-0.16147073869564943},{"x":39,"y":0.07481730106045208},{"x":40,"y":0.1963718573922713},{"x":41,"y":-0.05909009604450789},{"x":42,"y":-0.16516433875533731},{"x":43,"y":0.09485988777253589},{"x":44,"y":0.2121685016794648},{"x":45,"y":-0.04571464552797914},{"x":46,"y":-0.13168560181472763},{"x":47,"y":-0.012463668013549036},{"x":48,"y":0.16761365949589332},{"x":49,"y":0.059101222976395834}]}],"marks":[{"type":"line","from":{"data":"0585429e-c9cb-4cc4-8a66-dacca3e9363a"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"blue"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}},{"type":"line","from":{"data":"dc1ebe84-c3ed-41f5-9df8-e2831f15109f"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"red"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}}]},"value":"#gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"0585429e-c9cb-4cc4-8a66-dacca3e9363a\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"0585429e-c9cb-4cc4-8a66-dacca3e9363a\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"0585429e-c9cb-4cc4-8a66-dacca3e9363a\", :values ({:x 0, :y 1.4177} {:x 1, :y 0.76228} {:x 2, :y 1.079} {:x 3, :y -0.20891} {:x 4, :y -0.6084} {:x 5, :y 1.0057} {:x 6, :y 0.51189} {:x 7, :y -0.26634} {:x 8, :y -0.8711} {:x 9, :y -1.128} {:x 10, :y 0.030323} {:x 11, :y 1.5069} {:x 12, :y 0.41506} {:x 13, :y 0.23892} {:x 14, :y 0.54717} {:x 15, :y 0.92183} {:x 16, :y 1.1968} {:x 17, :y 1.6507} {:x 18, :y 0.90201} {:x 19, :y 1.6279} {:x 20, :y -0.15032} {:x 21, :y -0.87007} {:x 22, :y 0.89195} {:x 23, :y 0.53004} {:x 24, :y -0.44385} {:x 25, :y -0.62716} {:x 26, :y -0.99799} {:x 27, :y 0.24588} {:x 28, :y 1.5889} {:x 29, :y 0.51156} {:x 30, :y 0.19102} {:x 31, :y 0.50977} {:x 32, :y 0.9414} {:x 33, :y 1.2945} {:x 34, :y 1.5499} {:x 35, :y 1.3471} {:x 36, :y 1.8116} {:x 37, :y 0.34641} {:x 38, :y 0.10246} {:x 39, :y -0.37094} {:x 40, :y -0.30507} {:x 41, :y 1.0961} {:x 42, :y 0.67648} {:x 43, :y 0.99356} {:x 44, :y 0.012639} {:x 45, :y -0.3428} {:x 46, :y 0.77221} {:x 47, :y 0.58104} {:x 48, :y 0.9016} {:x 49, :y 0.58413})} {:name \"dc1ebe84-c3ed-41f5-9df8-e2831f15109f\", :values ({:x 0, :y 1.3387421635344436} {:x 1, :y 0.47872890986650246} {:x 2, :y -0.34134570045021984} {:x 3, :y -0.16034407351471264} {:x 4, :y 0.23278263418044728} {:x 5, :y 0.08937256267855709} {:x 6, :y -0.19924327655020221} {:x 7, :y -0.012050489565514359} {:x 8, :y 0.1440881803645548} {:x 9, :y 0.054289186459479735} {:x 10, :y -0.09192097157915488} {:x 11, :y 0.00705905140184483} {:x 12, :y 0.12375150517624058} {:x 13, :y 0.02440629257132847} {:x 14, :y -0.09111963326647114} {:x 15, :y 0.04548400618581666} {:x 16, :y 0.08906195165088954} {:x 17, :y 0.05475261757205917} {:x 18, :y -0.13442502250557192} {:x 19, :y -0.05066028402771136} {:x 20, :y 0.2092552812771087} {:x 21, :y -0.017591681548372852} {:x 22, :y -0.2392730147654148} {:x 23, :y 0.08860668830440048} {:x 24, :y 0.2005317917942492} {:x 25, :y -0.004314958816411618} {:x 26, :y -0.23042520244617012} {:x 27, :y 0.07060670823721525} {:x 28, :y 0.18565453792717046} {:x 29, :y -0.03559618378901941} {:x 30, :y -0.1740038594564115} {:x 31, :y 0.07817094628555893} {:x 32, :y 0.20731280618983428} {:x 33, :y -0.09500327490783189} {:x 34, :y -0.1366131547318104} {:x 35, :y 0.06309104086870862} {:x 36, :y 0.21343708472650297} {:x 37, :y -0.06358634686274946} {:x 38, :y -0.16147073869564943} {:x 39, :y 0.07481730106045208} {:x 40, :y 0.1963718573922713} {:x 41, :y -0.05909009604450789} {:x 42, :y -0.16516433875533731} {:x 43, :y 0.09485988777253589} {:x 44, :y 0.2121685016794648} {:x 45, :y -0.04571464552797914} {:x 46, :y -0.13168560181472763} {:x 47, :y -0.012463668013549036} {:x 48, :y 0.16761365949589332} {:x 49, :y 0.059101222976395834})}), :marks ({:type \"line\", :from {:data \"0585429e-c9cb-4cc4-8a66-dacca3e9363a\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"dc1ebe84-c3ed-41f5-9df8-e2831f15109f\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}}"},{"type":"vega","content":{"width":250,"height":154.51174926757812,"padding":{"top":10,"left":55,"bottom":40,"right":10},"scales":[{"name":"x","type":"linear","range":"width","zero":false,"domain":{"data":"2144c3e2-5202-4423-8967-37f950b745ab","field":"data.x"}},{"name":"y","type":"linear","range":"height","nice":true,"zero":false,"domain":{"data":"2144c3e2-5202-4423-8967-37f950b745ab","field":"data.y"}}],"axes":[{"type":"x","scale":"x","title":"Time step","titleOffset":30},{"type":"y","scale":"y"}],"data":[{"name":"2144c3e2-5202-4423-8967-37f950b745ab","values":[{"x":0,"y":-0.7683},{"x":1,"y":-0.37145},{"x":2,"y":-1.7936},{"x":3,"y":-0.067909},{"x":4,"y":-0.72467},{"x":5,"y":-0.7811},{"x":6,"y":0.010419},{"x":7,"y":-1.5865},{"x":8,"y":0.59129},{"x":9,"y":-0.99314},{"x":10,"y":-1.0829},{"x":11,"y":-0.17441},{"x":12,"y":-0.99972},{"x":13,"y":-0.8976},{"x":14,"y":-0.57532},{"x":15,"y":-0.92552},{"x":16,"y":-0.91328},{"x":17,"y":-1.1929},{"x":18,"y":-0.65438},{"x":19,"y":-1.657},{"x":20,"y":-0.025507},{"x":21,"y":-0.77629},{"x":22,"y":-0.7095},{"x":23,"y":0.1119},{"x":24,"y":-1.5459},{"x":25,"y":0.66598},{"x":26,"y":-1.02},{"x":27,"y":-0.87674},{"x":28,"y":-0.15425},{"x":29,"y":-1.1489},{"x":30,"y":-0.82843},{"x":31,"y":-0.64842},{"x":32,"y":-0.82055},{"x":33,"y":-1.0004},{"x":34,"y":-1.0683},{"x":35,"y":-0.93447},{"x":36,"y":-1.2922},{"x":37,"y":-0.35682},{"x":38,"y":-1.3844},{"x":39,"y":0.10704},{"x":40,"y":-0.93704},{"x":41,"y":-0.5491},{"x":42,"y":-0.474},{"x":43,"y":-1.585},{"x":44,"y":-0.22502},{"x":45,"y":-0.97659},{"x":46,"y":-0.33674},{"x":47,"y":-0.60478},{"x":48,"y":-1.2616},{"x":49,"y":-0.52966}]},{"name":"865fadbd-871b-42f1-856b-59e708b68949","values":[{"x":0,"y":-0.8293499331800976},{"x":1,"y":0.15783972105934674},{"x":2,"y":-1.5576963360259966},{"x":3,"y":0.8412054243521987},{"x":4,"y":0.4751856424350188},{"x":5,"y":-0.2784169338179098},{"x":6,"y":-0.7755770795952894},{"x":7,"y":0.4394589741106296},{"x":8,"y":-0.433310043442401},{"x":9,"y":-0.4295509442252003},{"x":10,"y":-0.4329839595433038},{"x":11,"y":-0.4175875663700794},{"x":12,"y":-0.482327270458645},{"x":13,"y":-0.3186602546389668},{"x":14,"y":-0.6833760650256998},{"x":15,"y":0.20650658963492013},{"x":16,"y":-0.8475328274390175},{"x":17,"y":0.5608551428829511},{"x":18,"y":-0.1335481273369694},{"x":19,"y":-0.9320501437967287},{"x":20,"y":0.7200609820915493},{"x":21,"y":0.29320874267842295},{"x":22,"y":-0.7260481043745287},{"x":23,"y":0.2862641818264845},{"x":24,"y":-0.710328855434687},{"x":25,"y":0.25683600299981063},{"x":26,"y":-0.7783978188650806},{"x":27,"y":0.4218060092739567},{"x":28,"y":-0.42253074084626985},{"x":29,"y":-0.49287519563428056},{"x":30,"y":-0.2935660797475224},{"x":31,"y":-0.7430616124821166},{"x":32,"y":0.326163937019353},{"x":33,"y":-0.7002947624051985},{"x":34,"y":0.26558384670361274},{"x":35,"y":-0.7874586217372538},{"x":36,"y":0.4450058832279967},{"x":37,"y":-0.4184397822746409},{"x":38,"y":-0.44862060924074565},{"x":39,"y":-0.4184477746266324},{"x":40,"y":-0.44105899113023916},{"x":41,"y":-0.43324936581538864},{"x":42,"y":-0.4193768487273578},{"x":43,"y":-0.4927825894288068},{"x":44,"y":-0.2731848663367741},{"x":45,"y":-0.7847318136616724},{"x":46,"y":0.45959200550100376},{"x":47,"y":-0.38940975657274657},{"x":48,"y":-0.5409666715909941},{"x":49,"y":-0.186823655959339}]}],"marks":[{"type":"line","from":{"data":"2144c3e2-5202-4423-8967-37f950b745ab"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"blue"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}},{"type":"line","from":{"data":"865fadbd-871b-42f1-856b-59e708b68949"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"red"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}}]},"value":"#gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"2144c3e2-5202-4423-8967-37f950b745ab\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"2144c3e2-5202-4423-8967-37f950b745ab\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"2144c3e2-5202-4423-8967-37f950b745ab\", :values ({:x 0, :y -0.7683} {:x 1, :y -0.37145} {:x 2, :y -1.7936} {:x 3, :y -0.067909} {:x 4, :y -0.72467} {:x 5, :y -0.7811} {:x 6, :y 0.010419} {:x 7, :y -1.5865} {:x 8, :y 0.59129} {:x 9, :y -0.99314} {:x 10, :y -1.0829} {:x 11, :y -0.17441} {:x 12, :y -0.99972} {:x 13, :y -0.8976} {:x 14, :y -0.57532} {:x 15, :y -0.92552} {:x 16, :y -0.91328} {:x 17, :y -1.1929} {:x 18, :y -0.65438} {:x 19, :y -1.657} {:x 20, :y -0.025507} {:x 21, :y -0.77629} {:x 22, :y -0.7095} {:x 23, :y 0.1119} {:x 24, :y -1.5459} {:x 25, :y 0.66598} {:x 26, :y -1.02} {:x 27, :y -0.87674} {:x 28, :y -0.15425} {:x 29, :y -1.1489} {:x 30, :y -0.82843} {:x 31, :y -0.64842} {:x 32, :y -0.82055} {:x 33, :y -1.0004} {:x 34, :y -1.0683} {:x 35, :y -0.93447} {:x 36, :y -1.2922} {:x 37, :y -0.35682} {:x 38, :y -1.3844} {:x 39, :y 0.10704} {:x 40, :y -0.93704} {:x 41, :y -0.5491} {:x 42, :y -0.474} {:x 43, :y -1.585} {:x 44, :y -0.22502} {:x 45, :y -0.97659} {:x 46, :y -0.33674} {:x 47, :y -0.60478} {:x 48, :y -1.2616} {:x 49, :y -0.52966})} {:name \"865fadbd-871b-42f1-856b-59e708b68949\", :values ({:x 0, :y -0.8293499331800976} {:x 1, :y 0.15783972105934674} {:x 2, :y -1.5576963360259966} {:x 3, :y 0.8412054243521987} {:x 4, :y 0.4751856424350188} {:x 5, :y -0.2784169338179098} {:x 6, :y -0.7755770795952894} {:x 7, :y 0.4394589741106296} {:x 8, :y -0.433310043442401} {:x 9, :y -0.4295509442252003} {:x 10, :y -0.4329839595433038} {:x 11, :y -0.4175875663700794} {:x 12, :y -0.482327270458645} {:x 13, :y -0.3186602546389668} {:x 14, :y -0.6833760650256998} {:x 15, :y 0.20650658963492013} {:x 16, :y -0.8475328274390175} {:x 17, :y 0.5608551428829511} {:x 18, :y -0.1335481273369694} {:x 19, :y -0.9320501437967287} {:x 20, :y 0.7200609820915493} {:x 21, :y 0.29320874267842295} {:x 22, :y -0.7260481043745287} {:x 23, :y 0.2862641818264845} {:x 24, :y -0.710328855434687} {:x 25, :y 0.25683600299981063} {:x 26, :y -0.7783978188650806} {:x 27, :y 0.4218060092739567} {:x 28, :y -0.42253074084626985} {:x 29, :y -0.49287519563428056} {:x 30, :y -0.2935660797475224} {:x 31, :y -0.7430616124821166} {:x 32, :y 0.326163937019353} {:x 33, :y -0.7002947624051985} {:x 34, :y 0.26558384670361274} {:x 35, :y -0.7874586217372538} {:x 36, :y 0.4450058832279967} {:x 37, :y -0.4184397822746409} {:x 38, :y -0.44862060924074565} {:x 39, :y -0.4184477746266324} {:x 40, :y -0.44105899113023916} {:x 41, :y -0.43324936581538864} {:x 42, :y -0.4193768487273578} {:x 43, :y -0.4927825894288068} {:x 44, :y -0.2731848663367741} {:x 45, :y -0.7847318136616724} {:x 46, :y 0.45959200550100376} {:x 47, :y -0.38940975657274657} {:x 48, :y -0.5409666715909941} {:x 49, :y -0.186823655959339})}), :marks ({:type \"line\", :from {:data \"2144c3e2-5202-4423-8967-37f950b745ab\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"865fadbd-871b-42f1-856b-59e708b68949\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}}"},{"type":"vega","content":{"width":250,"height":154.51174926757812,"padding":{"top":10,"left":55,"bottom":40,"right":10},"scales":[{"name":"x","type":"linear","range":"width","zero":false,"domain":{"data":"5be9bc10-a9f5-4a09-ac0f-3f3e48c6baf0","field":"data.x"}},{"name":"y","type":"linear","range":"height","nice":true,"zero":false,"domain":{"data":"5be9bc10-a9f5-4a09-ac0f-3f3e48c6baf0","field":"data.y"}}],"axes":[{"type":"x","scale":"x","title":"Time step","titleOffset":30},{"type":"y","scale":"y"}],"data":[{"name":"5be9bc10-a9f5-4a09-ac0f-3f3e48c6baf0","values":[{"x":0,"y":-0.23752},{"x":1,"y":0.98831},{"x":2,"y":0.69057},{"x":3,"y":0.88147},{"x":4,"y":-0.2074},{"x":5,"y":-0.57156},{"x":6,"y":0.84455},{"x":7,"y":0.48982},{"x":8,"y":-0.2632},{"x":9,"y":-0.76504},{"x":10,"y":-0.90355},{"x":11,"y":0.030318},{"x":12,"y":0.99796},{"x":13,"y":0.40324},{"x":14,"y":0.23665},{"x":15,"y":0.52027},{"x":16,"y":0.79671},{"x":17,"y":0.93087},{"x":18,"y":0.99681},{"x":19,"y":0.78457},{"x":20,"y":0.99837},{"x":21,"y":-0.14975},{"x":22,"y":-0.76437},{"x":23,"y":0.7783},{"x":24,"y":0.50557},{"x":25,"y":-0.42942},{"x":26,"y":-0.58685},{"x":27,"y":-0.84038},{"x":28,"y":0.24341},{"x":29,"y":0.99984},{"x":30,"y":0.48954},{"x":31,"y":0.18986},{"x":32,"y":0.48798},{"x":33,"y":0.80838},{"x":34,"y":0.96206},{"x":35,"y":0.99978},{"x":36,"y":0.97508},{"x":37,"y":0.97114},{"x":38,"y":0.33953},{"x":39,"y":0.10228},{"x":40,"y":-0.36249},{"x":41,"y":-0.30036},{"x":42,"y":0.88944},{"x":43,"y":0.62605},{"x":44,"y":0.83797},{"x":45,"y":0.012639},{"x":46,"y":-0.33613},{"x":47,"y":0.69772},{"x":48,"y":0.54889},{"x":49,"y":0.78432}]},{"name":"0195ec20-0c43-422c-85af-32be32d9cac9","values":[{"x":0,"y":0.4192690877644614},{"x":1,"y":0.9626168294534553},{"x":2,"y":0.45783964474985855},{"x":3,"y":-0.33319908698290224},{"x":4,"y":-0.16543985092235725},{"x":5,"y":0.23174379560072694},{"x":6,"y":0.0864978865574241},{"x":7,"y":-0.18750400543506018},{"x":8,"y":-0.017856706800764836},{"x":9,"y":0.14194084638225915},{"x":10,"y":0.038191024072481386},{"x":11,"y":-0.09619944524152588},{"x":12,"y":0.007988896948398057},{"x":13,"y":0.11698736311769764},{"x":14,"y":0.00737560427361483},{"x":15,"y":-0.10965304825246867},{"x":16,"y":0.038929635307386454},{"x":17,"y":0.08573166846386077},{"x":18,"y":0.05844168696508747},{"x":19,"y":-0.12883554878690928},{"x":20,"y":-0.05968487877420587},{"x":21,"y":0.21873333614090248},{"x":22,"y":-0.0341934170494822},{"x":23,"y":-0.24482360450421117},{"x":24,"y":0.08093118938036865},{"x":25,"y":0.1942099290176708},{"x":26,"y":0.010121628322143642},{"x":27,"y":-0.22934799010053153},{"x":28,"y":0.08266886934267755},{"x":29,"y":0.19558898180933068},{"x":30,"y":-0.036627085429146354},{"x":31,"y":-0.1482140722222245},{"x":32,"y":0.06904445179876972},{"x":33,"y":0.20222021117991432},{"x":34,"y":-0.0997077367289781},{"x":35,"y":-0.14307401957597085},{"x":36,"y":0.05525733972514405},{"x":37,"y":0.19961645986275653},{"x":38,"y":-0.04997464866937276},{"x":39,"y":-0.1619084904817675},{"x":40,"y":0.08536360455520432},{"x":41,"y":0.21208907414690614},{"x":42,"y":-0.0698212915579997},{"x":43,"y":-0.15688300010460438},{"x":44,"y":0.08798836315246568},{"x":45,"y":0.1964783894535316},{"x":46,"y":-0.03264494060941437},{"x":47,"y":-0.13023782529080002},{"x":48,"y":-0.005481290037255693},{"x":49,"y":0.14546768768682702}]}],"marks":[{"type":"line","from":{"data":"5be9bc10-a9f5-4a09-ac0f-3f3e48c6baf0"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"blue"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}},{"type":"line","from":{"data":"0195ec20-0c43-422c-85af-32be32d9cac9"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"red"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}}]},"value":"#gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"5be9bc10-a9f5-4a09-ac0f-3f3e48c6baf0\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"5be9bc10-a9f5-4a09-ac0f-3f3e48c6baf0\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"5be9bc10-a9f5-4a09-ac0f-3f3e48c6baf0\", :values ({:x 0, :y -0.23752} {:x 1, :y 0.98831} {:x 2, :y 0.69057} {:x 3, :y 0.88147} {:x 4, :y -0.2074} {:x 5, :y -0.57156} {:x 6, :y 0.84455} {:x 7, :y 0.48982} {:x 8, :y -0.2632} {:x 9, :y -0.76504} {:x 10, :y -0.90355} {:x 11, :y 0.030318} {:x 12, :y 0.99796} {:x 13, :y 0.40324} {:x 14, :y 0.23665} {:x 15, :y 0.52027} {:x 16, :y 0.79671} {:x 17, :y 0.93087} {:x 18, :y 0.99681} {:x 19, :y 0.78457} {:x 20, :y 0.99837} {:x 21, :y -0.14975} {:x 22, :y -0.76437} {:x 23, :y 0.7783} {:x 24, :y 0.50557} {:x 25, :y -0.42942} {:x 26, :y -0.58685} {:x 27, :y -0.84038} {:x 28, :y 0.24341} {:x 29, :y 0.99984} {:x 30, :y 0.48954} {:x 31, :y 0.18986} {:x 32, :y 0.48798} {:x 33, :y 0.80838} {:x 34, :y 0.96206} {:x 35, :y 0.99978} {:x 36, :y 0.97508} {:x 37, :y 0.97114} {:x 38, :y 0.33953} {:x 39, :y 0.10228} {:x 40, :y -0.36249} {:x 41, :y -0.30036} {:x 42, :y 0.88944} {:x 43, :y 0.62605} {:x 44, :y 0.83797} {:x 45, :y 0.012639} {:x 46, :y -0.33613} {:x 47, :y 0.69772} {:x 48, :y 0.54889} {:x 49, :y 0.78432})} {:name \"0195ec20-0c43-422c-85af-32be32d9cac9\", :values ({:x 0, :y 0.4192690877644614} {:x 1, :y 0.9626168294534553} {:x 2, :y 0.45783964474985855} {:x 3, :y -0.33319908698290224} {:x 4, :y -0.16543985092235725} {:x 5, :y 0.23174379560072694} {:x 6, :y 0.0864978865574241} {:x 7, :y -0.18750400543506018} {:x 8, :y -0.017856706800764836} {:x 9, :y 0.14194084638225915} {:x 10, :y 0.038191024072481386} {:x 11, :y -0.09619944524152588} {:x 12, :y 0.007988896948398057} {:x 13, :y 0.11698736311769764} {:x 14, :y 0.00737560427361483} {:x 15, :y -0.10965304825246867} {:x 16, :y 0.038929635307386454} {:x 17, :y 0.08573166846386077} {:x 18, :y 0.05844168696508747} {:x 19, :y -0.12883554878690928} {:x 20, :y -0.05968487877420587} {:x 21, :y 0.21873333614090248} {:x 22, :y -0.0341934170494822} {:x 23, :y -0.24482360450421117} {:x 24, :y 0.08093118938036865} {:x 25, :y 0.1942099290176708} {:x 26, :y 0.010121628322143642} {:x 27, :y -0.22934799010053153} {:x 28, :y 0.08266886934267755} {:x 29, :y 0.19558898180933068} {:x 30, :y -0.036627085429146354} {:x 31, :y -0.1482140722222245} {:x 32, :y 0.06904445179876972} {:x 33, :y 0.20222021117991432} {:x 34, :y -0.0997077367289781} {:x 35, :y -0.14307401957597085} {:x 36, :y 0.05525733972514405} {:x 37, :y 0.19961645986275653} {:x 38, :y -0.04997464866937276} {:x 39, :y -0.1619084904817675} {:x 40, :y 0.08536360455520432} {:x 41, :y 0.21208907414690614} {:x 42, :y -0.0698212915579997} {:x 43, :y -0.15688300010460438} {:x 44, :y 0.08798836315246568} {:x 45, :y 0.1964783894535316} {:x 46, :y -0.03264494060941437} {:x 47, :y -0.13023782529080002} {:x 48, :y -0.005481290037255693} {:x 49, :y 0.14546768768682702})}), :marks ({:type \"line\", :from {:data \"5be9bc10-a9f5-4a09-ac0f-3f3e48c6baf0\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"0195ec20-0c43-422c-85af-32be32d9cac9\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}}"}],"value":"[#gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"0585429e-c9cb-4cc4-8a66-dacca3e9363a\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"0585429e-c9cb-4cc4-8a66-dacca3e9363a\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"0585429e-c9cb-4cc4-8a66-dacca3e9363a\", :values ({:x 0, :y 1.4177} {:x 1, :y 0.76228} {:x 2, :y 1.079} {:x 3, :y -0.20891} {:x 4, :y -0.6084} {:x 5, :y 1.0057} {:x 6, :y 0.51189} {:x 7, :y -0.26634} {:x 8, :y -0.8711} {:x 9, :y -1.128} {:x 10, :y 0.030323} {:x 11, :y 1.5069} {:x 12, :y 0.41506} {:x 13, :y 0.23892} {:x 14, :y 0.54717} {:x 15, :y 0.92183} {:x 16, :y 1.1968} {:x 17, :y 1.6507} {:x 18, :y 0.90201} {:x 19, :y 1.6279} {:x 20, :y -0.15032} {:x 21, :y -0.87007} {:x 22, :y 0.89195} {:x 23, :y 0.53004} {:x 24, :y -0.44385} {:x 25, :y -0.62716} {:x 26, :y -0.99799} {:x 27, :y 0.24588} {:x 28, :y 1.5889} {:x 29, :y 0.51156} {:x 30, :y 0.19102} {:x 31, :y 0.50977} {:x 32, :y 0.9414} {:x 33, :y 1.2945} {:x 34, :y 1.5499} {:x 35, :y 1.3471} {:x 36, :y 1.8116} {:x 37, :y 0.34641} {:x 38, :y 0.10246} {:x 39, :y -0.37094} {:x 40, :y -0.30507} {:x 41, :y 1.0961} {:x 42, :y 0.67648} {:x 43, :y 0.99356} {:x 44, :y 0.012639} {:x 45, :y -0.3428} {:x 46, :y 0.77221} {:x 47, :y 0.58104} {:x 48, :y 0.9016} {:x 49, :y 0.58413})} {:name \"dc1ebe84-c3ed-41f5-9df8-e2831f15109f\", :values ({:x 0, :y 1.3387421635344436} {:x 1, :y 0.47872890986650246} {:x 2, :y -0.34134570045021984} {:x 3, :y -0.16034407351471264} {:x 4, :y 0.23278263418044728} {:x 5, :y 0.08937256267855709} {:x 6, :y -0.19924327655020221} {:x 7, :y -0.012050489565514359} {:x 8, :y 0.1440881803645548} {:x 9, :y 0.054289186459479735} {:x 10, :y -0.09192097157915488} {:x 11, :y 0.00705905140184483} {:x 12, :y 0.12375150517624058} {:x 13, :y 0.02440629257132847} {:x 14, :y -0.09111963326647114} {:x 15, :y 0.04548400618581666} {:x 16, :y 0.08906195165088954} {:x 17, :y 0.05475261757205917} {:x 18, :y -0.13442502250557192} {:x 19, :y -0.05066028402771136} {:x 20, :y 0.2092552812771087} {:x 21, :y -0.017591681548372852} {:x 22, :y -0.2392730147654148} {:x 23, :y 0.08860668830440048} {:x 24, :y 0.2005317917942492} {:x 25, :y -0.004314958816411618} {:x 26, :y -0.23042520244617012} {:x 27, :y 0.07060670823721525} {:x 28, :y 0.18565453792717046} {:x 29, :y -0.03559618378901941} {:x 30, :y -0.1740038594564115} {:x 31, :y 0.07817094628555893} {:x 32, :y 0.20731280618983428} {:x 33, :y -0.09500327490783189} {:x 34, :y -0.1366131547318104} {:x 35, :y 0.06309104086870862} {:x 36, :y 0.21343708472650297} {:x 37, :y -0.06358634686274946} {:x 38, :y -0.16147073869564943} {:x 39, :y 0.07481730106045208} {:x 40, :y 0.1963718573922713} {:x 41, :y -0.05909009604450789} {:x 42, :y -0.16516433875533731} {:x 43, :y 0.09485988777253589} {:x 44, :y 0.2121685016794648} {:x 45, :y -0.04571464552797914} {:x 46, :y -0.13168560181472763} {:x 47, :y -0.012463668013549036} {:x 48, :y 0.16761365949589332} {:x 49, :y 0.059101222976395834})}), :marks ({:type \"line\", :from {:data \"0585429e-c9cb-4cc4-8a66-dacca3e9363a\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"dc1ebe84-c3ed-41f5-9df8-e2831f15109f\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}} #gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"2144c3e2-5202-4423-8967-37f950b745ab\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"2144c3e2-5202-4423-8967-37f950b745ab\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"2144c3e2-5202-4423-8967-37f950b745ab\", :values ({:x 0, :y -0.7683} {:x 1, :y -0.37145} {:x 2, :y -1.7936} {:x 3, :y -0.067909} {:x 4, :y -0.72467} {:x 5, :y -0.7811} {:x 6, :y 0.010419} {:x 7, :y -1.5865} {:x 8, :y 0.59129} {:x 9, :y -0.99314} {:x 10, :y -1.0829} {:x 11, :y -0.17441} {:x 12, :y -0.99972} {:x 13, :y -0.8976} {:x 14, :y -0.57532} {:x 15, :y -0.92552} {:x 16, :y -0.91328} {:x 17, :y -1.1929} {:x 18, :y -0.65438} {:x 19, :y -1.657} {:x 20, :y -0.025507} {:x 21, :y -0.77629} {:x 22, :y -0.7095} {:x 23, :y 0.1119} {:x 24, :y -1.5459} {:x 25, :y 0.66598} {:x 26, :y -1.02} {:x 27, :y -0.87674} {:x 28, :y -0.15425} {:x 29, :y -1.1489} {:x 30, :y -0.82843} {:x 31, :y -0.64842} {:x 32, :y -0.82055} {:x 33, :y -1.0004} {:x 34, :y -1.0683} {:x 35, :y -0.93447} {:x 36, :y -1.2922} {:x 37, :y -0.35682} {:x 38, :y -1.3844} {:x 39, :y 0.10704} {:x 40, :y -0.93704} {:x 41, :y -0.5491} {:x 42, :y -0.474} {:x 43, :y -1.585} {:x 44, :y -0.22502} {:x 45, :y -0.97659} {:x 46, :y -0.33674} {:x 47, :y -0.60478} {:x 48, :y -1.2616} {:x 49, :y -0.52966})} {:name \"865fadbd-871b-42f1-856b-59e708b68949\", :values ({:x 0, :y -0.8293499331800976} {:x 1, :y 0.15783972105934674} {:x 2, :y -1.5576963360259966} {:x 3, :y 0.8412054243521987} {:x 4, :y 0.4751856424350188} {:x 5, :y -0.2784169338179098} {:x 6, :y -0.7755770795952894} {:x 7, :y 0.4394589741106296} {:x 8, :y -0.433310043442401} {:x 9, :y -0.4295509442252003} {:x 10, :y -0.4329839595433038} {:x 11, :y -0.4175875663700794} {:x 12, :y -0.482327270458645} {:x 13, :y -0.3186602546389668} {:x 14, :y -0.6833760650256998} {:x 15, :y 0.20650658963492013} {:x 16, :y -0.8475328274390175} {:x 17, :y 0.5608551428829511} {:x 18, :y -0.1335481273369694} {:x 19, :y -0.9320501437967287} {:x 20, :y 0.7200609820915493} {:x 21, :y 0.29320874267842295} {:x 22, :y -0.7260481043745287} {:x 23, :y 0.2862641818264845} {:x 24, :y -0.710328855434687} {:x 25, :y 0.25683600299981063} {:x 26, :y -0.7783978188650806} {:x 27, :y 0.4218060092739567} {:x 28, :y -0.42253074084626985} {:x 29, :y -0.49287519563428056} {:x 30, :y -0.2935660797475224} {:x 31, :y -0.7430616124821166} {:x 32, :y 0.326163937019353} {:x 33, :y -0.7002947624051985} {:x 34, :y 0.26558384670361274} {:x 35, :y -0.7874586217372538} {:x 36, :y 0.4450058832279967} {:x 37, :y -0.4184397822746409} {:x 38, :y -0.44862060924074565} {:x 39, :y -0.4184477746266324} {:x 40, :y -0.44105899113023916} {:x 41, :y -0.43324936581538864} {:x 42, :y -0.4193768487273578} {:x 43, :y -0.4927825894288068} {:x 44, :y -0.2731848663367741} {:x 45, :y -0.7847318136616724} {:x 46, :y 0.45959200550100376} {:x 47, :y -0.38940975657274657} {:x 48, :y -0.5409666715909941} {:x 49, :y -0.186823655959339})}), :marks ({:type \"line\", :from {:data \"2144c3e2-5202-4423-8967-37f950b745ab\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"865fadbd-871b-42f1-856b-59e708b68949\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}} #gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"5be9bc10-a9f5-4a09-ac0f-3f3e48c6baf0\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"5be9bc10-a9f5-4a09-ac0f-3f3e48c6baf0\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"5be9bc10-a9f5-4a09-ac0f-3f3e48c6baf0\", :values ({:x 0, :y -0.23752} {:x 1, :y 0.98831} {:x 2, :y 0.69057} {:x 3, :y 0.88147} {:x 4, :y -0.2074} {:x 5, :y -0.57156} {:x 6, :y 0.84455} {:x 7, :y 0.48982} {:x 8, :y -0.2632} {:x 9, :y -0.76504} {:x 10, :y -0.90355} {:x 11, :y 0.030318} {:x 12, :y 0.99796} {:x 13, :y 0.40324} {:x 14, :y 0.23665} {:x 15, :y 0.52027} {:x 16, :y 0.79671} {:x 17, :y 0.93087} {:x 18, :y 0.99681} {:x 19, :y 0.78457} {:x 20, :y 0.99837} {:x 21, :y -0.14975} {:x 22, :y -0.76437} {:x 23, :y 0.7783} {:x 24, :y 0.50557} {:x 25, :y -0.42942} {:x 26, :y -0.58685} {:x 27, :y -0.84038} {:x 28, :y 0.24341} {:x 29, :y 0.99984} {:x 30, :y 0.48954} {:x 31, :y 0.18986} {:x 32, :y 0.48798} {:x 33, :y 0.80838} {:x 34, :y 0.96206} {:x 35, :y 0.99978} {:x 36, :y 0.97508} {:x 37, :y 0.97114} {:x 38, :y 0.33953} {:x 39, :y 0.10228} {:x 40, :y -0.36249} {:x 41, :y -0.30036} {:x 42, :y 0.88944} {:x 43, :y 0.62605} {:x 44, :y 0.83797} {:x 45, :y 0.012639} {:x 46, :y -0.33613} {:x 47, :y 0.69772} {:x 48, :y 0.54889} {:x 49, :y 0.78432})} {:name \"0195ec20-0c43-422c-85af-32be32d9cac9\", :values ({:x 0, :y 0.4192690877644614} {:x 1, :y 0.9626168294534553} {:x 2, :y 0.45783964474985855} {:x 3, :y -0.33319908698290224} {:x 4, :y -0.16543985092235725} {:x 5, :y 0.23174379560072694} {:x 6, :y 0.0864978865574241} {:x 7, :y -0.18750400543506018} {:x 8, :y -0.017856706800764836} {:x 9, :y 0.14194084638225915} {:x 10, :y 0.038191024072481386} {:x 11, :y -0.09619944524152588} {:x 12, :y 0.007988896948398057} {:x 13, :y 0.11698736311769764} {:x 14, :y 0.00737560427361483} {:x 15, :y -0.10965304825246867} {:x 16, :y 0.038929635307386454} {:x 17, :y 0.08573166846386077} {:x 18, :y 0.05844168696508747} {:x 19, :y -0.12883554878690928} {:x 20, :y -0.05968487877420587} {:x 21, :y 0.21873333614090248} {:x 22, :y -0.0341934170494822} {:x 23, :y -0.24482360450421117} {:x 24, :y 0.08093118938036865} {:x 25, :y 0.1942099290176708} {:x 26, :y 0.010121628322143642} {:x 27, :y -0.22934799010053153} {:x 28, :y 0.08266886934267755} {:x 29, :y 0.19558898180933068} {:x 30, :y -0.036627085429146354} {:x 31, :y -0.1482140722222245} {:x 32, :y 0.06904445179876972} {:x 33, :y 0.20222021117991432} {:x 34, :y -0.0997077367289781} {:x 35, :y -0.14307401957597085} {:x 36, :y 0.05525733972514405} {:x 37, :y 0.19961645986275653} {:x 38, :y -0.04997464866937276} {:x 39, :y -0.1619084904817675} {:x 40, :y 0.08536360455520432} {:x 41, :y 0.21208907414690614} {:x 42, :y -0.0698212915579997} {:x 43, :y -0.15688300010460438} {:x 44, :y 0.08798836315246568} {:x 45, :y 0.1964783894535316} {:x 46, :y -0.03264494060941437} {:x 47, :y -0.13023782529080002} {:x 48, :y -0.005481290037255693} {:x 49, :y 0.14546768768682702})}), :marks ({:type \"line\", :from {:data \"5be9bc10-a9f5-4a09-ac0f-3f3e48c6baf0\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"0195ec20-0c43-422c-85af-32be32d9cac9\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}}]"}
;; <=

;; **
;;; Ground truth (blue) and inferred (red) trajectories. Left to right: dimensions 1 to 3. After **20 iterations of BOPP**.
;; **

;; @@
(mapv 
  #(plot/compose
     (plot/list-plot %1
                     :joined true 
                     :color (first plot-colors)
                     :plot-size 250
                     :x-title "Time step")
     (plot/list-plot %2
                     :joined true 
                     :color (second plot-colors)))
  ground-truth-trajectory
  (nth inferred-trajectories 19))
;; @@
;; =>
;;; {"type":"list-like","open":"<span class='clj-vector'>[</span>","close":"<span class='clj-vector'>]</span>","separator":" ","items":[{"type":"vega","content":{"width":250,"height":154.51174926757812,"padding":{"top":10,"left":55,"bottom":40,"right":10},"scales":[{"name":"x","type":"linear","range":"width","zero":false,"domain":{"data":"32c07c23-a435-45b6-9bf9-abec1ba19a41","field":"data.x"}},{"name":"y","type":"linear","range":"height","nice":true,"zero":false,"domain":{"data":"32c07c23-a435-45b6-9bf9-abec1ba19a41","field":"data.y"}}],"axes":[{"type":"x","scale":"x","title":"Time step","titleOffset":30},{"type":"y","scale":"y"}],"data":[{"name":"32c07c23-a435-45b6-9bf9-abec1ba19a41","values":[{"x":0,"y":1.4177},{"x":1,"y":0.76228},{"x":2,"y":1.079},{"x":3,"y":-0.20891},{"x":4,"y":-0.6084},{"x":5,"y":1.0057},{"x":6,"y":0.51189},{"x":7,"y":-0.26634},{"x":8,"y":-0.8711},{"x":9,"y":-1.128},{"x":10,"y":0.030323},{"x":11,"y":1.5069},{"x":12,"y":0.41506},{"x":13,"y":0.23892},{"x":14,"y":0.54717},{"x":15,"y":0.92183},{"x":16,"y":1.1968},{"x":17,"y":1.6507},{"x":18,"y":0.90201},{"x":19,"y":1.6279},{"x":20,"y":-0.15032},{"x":21,"y":-0.87007},{"x":22,"y":0.89195},{"x":23,"y":0.53004},{"x":24,"y":-0.44385},{"x":25,"y":-0.62716},{"x":26,"y":-0.99799},{"x":27,"y":0.24588},{"x":28,"y":1.5889},{"x":29,"y":0.51156},{"x":30,"y":0.19102},{"x":31,"y":0.50977},{"x":32,"y":0.9414},{"x":33,"y":1.2945},{"x":34,"y":1.5499},{"x":35,"y":1.3471},{"x":36,"y":1.8116},{"x":37,"y":0.34641},{"x":38,"y":0.10246},{"x":39,"y":-0.37094},{"x":40,"y":-0.30507},{"x":41,"y":1.0961},{"x":42,"y":0.67648},{"x":43,"y":0.99356},{"x":44,"y":0.012639},{"x":45,"y":-0.3428},{"x":46,"y":0.77221},{"x":47,"y":0.58104},{"x":48,"y":0.9016},{"x":49,"y":0.58413}]},{"name":"573e36c4-f2f4-4afb-9bcb-6f4d5072396c","values":[{"x":0,"y":1.3402538531627508},{"x":1,"y":1.0322002619864143},{"x":2,"y":1.8014341933167113},{"x":3,"y":-0.16000458313615792},{"x":4,"y":-0.7744483571217643},{"x":5,"y":0.9363745319764447},{"x":6,"y":0.5254895921062439},{"x":7,"y":-0.29915283938028475},{"x":8,"y":-0.5038928058727293},{"x":9,"y":-0.6955922618142655},{"x":10,"y":0.6898091803077965},{"x":11,"y":0.9016269297091006},{"x":12,"y":0.557289695078088},{"x":13,"y":-0.2984692001144609},{"x":14,"y":0.32091215648940197},{"x":15,"y":1.1914191725037722},{"x":16,"y":1.2295436136266658},{"x":17,"y":1.6593034716783146},{"x":18,"y":1.110223671772022},{"x":19,"y":1.9275870897723504},{"x":20,"y":-0.25857193351522606},{"x":21,"y":-0.7170811459608197},{"x":22,"y":0.9169626198614169},{"x":23,"y":0.5224845031148656},{"x":24,"y":-0.49460789861013116},{"x":25,"y":-0.1857460198076708},{"x":26,"y":-0.4231543651316923},{"x":27,"y":0.9685359407715033},{"x":28,"y":0.5723638113647693},{"x":29,"y":0.6439790481989783},{"x":30,"y":0.07648367501862453},{"x":31,"y":-0.009815628541591737},{"x":32,"y":0.7953490708998123},{"x":33,"y":0.7743124109915819},{"x":34,"y":1.207594294421123},{"x":35,"y":1.352987814356697},{"x":36,"y":1.8590954334985326},{"x":37,"y":0.4543099495878354},{"x":38,"y":0.35608643974543747},{"x":39,"y":-0.24257402352852445},{"x":40,"y":-0.0386439876811051},{"x":41,"y":1.2391330793554265},{"x":42,"y":0.903430610994317},{"x":43,"y":1.6197532790143159},{"x":44,"y":0.2611887771032473},{"x":45,"y":-0.29700438654409667},{"x":46,"y":0.1841301329648706},{"x":47,"y":0.2128680631461865},{"x":48,"y":0.7816724284615412},{"x":49,"y":1.007302031506389}]}],"marks":[{"type":"line","from":{"data":"32c07c23-a435-45b6-9bf9-abec1ba19a41"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"blue"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}},{"type":"line","from":{"data":"573e36c4-f2f4-4afb-9bcb-6f4d5072396c"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"red"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}}]},"value":"#gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"32c07c23-a435-45b6-9bf9-abec1ba19a41\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"32c07c23-a435-45b6-9bf9-abec1ba19a41\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"32c07c23-a435-45b6-9bf9-abec1ba19a41\", :values ({:x 0, :y 1.4177} {:x 1, :y 0.76228} {:x 2, :y 1.079} {:x 3, :y -0.20891} {:x 4, :y -0.6084} {:x 5, :y 1.0057} {:x 6, :y 0.51189} {:x 7, :y -0.26634} {:x 8, :y -0.8711} {:x 9, :y -1.128} {:x 10, :y 0.030323} {:x 11, :y 1.5069} {:x 12, :y 0.41506} {:x 13, :y 0.23892} {:x 14, :y 0.54717} {:x 15, :y 0.92183} {:x 16, :y 1.1968} {:x 17, :y 1.6507} {:x 18, :y 0.90201} {:x 19, :y 1.6279} {:x 20, :y -0.15032} {:x 21, :y -0.87007} {:x 22, :y 0.89195} {:x 23, :y 0.53004} {:x 24, :y -0.44385} {:x 25, :y -0.62716} {:x 26, :y -0.99799} {:x 27, :y 0.24588} {:x 28, :y 1.5889} {:x 29, :y 0.51156} {:x 30, :y 0.19102} {:x 31, :y 0.50977} {:x 32, :y 0.9414} {:x 33, :y 1.2945} {:x 34, :y 1.5499} {:x 35, :y 1.3471} {:x 36, :y 1.8116} {:x 37, :y 0.34641} {:x 38, :y 0.10246} {:x 39, :y -0.37094} {:x 40, :y -0.30507} {:x 41, :y 1.0961} {:x 42, :y 0.67648} {:x 43, :y 0.99356} {:x 44, :y 0.012639} {:x 45, :y -0.3428} {:x 46, :y 0.77221} {:x 47, :y 0.58104} {:x 48, :y 0.9016} {:x 49, :y 0.58413})} {:name \"573e36c4-f2f4-4afb-9bcb-6f4d5072396c\", :values ({:x 0, :y 1.3402538531627508} {:x 1, :y 1.0322002619864143} {:x 2, :y 1.8014341933167113} {:x 3, :y -0.16000458313615792} {:x 4, :y -0.7744483571217643} {:x 5, :y 0.9363745319764447} {:x 6, :y 0.5254895921062439} {:x 7, :y -0.29915283938028475} {:x 8, :y -0.5038928058727293} {:x 9, :y -0.6955922618142655} {:x 10, :y 0.6898091803077965} {:x 11, :y 0.9016269297091006} {:x 12, :y 0.557289695078088} {:x 13, :y -0.2984692001144609} {:x 14, :y 0.32091215648940197} {:x 15, :y 1.1914191725037722} {:x 16, :y 1.2295436136266658} {:x 17, :y 1.6593034716783146} {:x 18, :y 1.110223671772022} {:x 19, :y 1.9275870897723504} {:x 20, :y -0.25857193351522606} {:x 21, :y -0.7170811459608197} {:x 22, :y 0.9169626198614169} {:x 23, :y 0.5224845031148656} {:x 24, :y -0.49460789861013116} {:x 25, :y -0.1857460198076708} {:x 26, :y -0.4231543651316923} {:x 27, :y 0.9685359407715033} {:x 28, :y 0.5723638113647693} {:x 29, :y 0.6439790481989783} {:x 30, :y 0.07648367501862453} {:x 31, :y -0.009815628541591737} {:x 32, :y 0.7953490708998123} {:x 33, :y 0.7743124109915819} {:x 34, :y 1.207594294421123} {:x 35, :y 1.352987814356697} {:x 36, :y 1.8590954334985326} {:x 37, :y 0.4543099495878354} {:x 38, :y 0.35608643974543747} {:x 39, :y -0.24257402352852445} {:x 40, :y -0.0386439876811051} {:x 41, :y 1.2391330793554265} {:x 42, :y 0.903430610994317} {:x 43, :y 1.6197532790143159} {:x 44, :y 0.2611887771032473} {:x 45, :y -0.29700438654409667} {:x 46, :y 0.1841301329648706} {:x 47, :y 0.2128680631461865} {:x 48, :y 0.7816724284615412} {:x 49, :y 1.007302031506389})}), :marks ({:type \"line\", :from {:data \"32c07c23-a435-45b6-9bf9-abec1ba19a41\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"573e36c4-f2f4-4afb-9bcb-6f4d5072396c\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}}"},{"type":"vega","content":{"width":250,"height":154.51174926757812,"padding":{"top":10,"left":55,"bottom":40,"right":10},"scales":[{"name":"x","type":"linear","range":"width","zero":false,"domain":{"data":"63c708e6-946d-4d57-8874-b6722b2ff366","field":"data.x"}},{"name":"y","type":"linear","range":"height","nice":true,"zero":false,"domain":{"data":"63c708e6-946d-4d57-8874-b6722b2ff366","field":"data.y"}}],"axes":[{"type":"x","scale":"x","title":"Time step","titleOffset":30},{"type":"y","scale":"y"}],"data":[{"name":"63c708e6-946d-4d57-8874-b6722b2ff366","values":[{"x":0,"y":-0.7683},{"x":1,"y":-0.37145},{"x":2,"y":-1.7936},{"x":3,"y":-0.067909},{"x":4,"y":-0.72467},{"x":5,"y":-0.7811},{"x":6,"y":0.010419},{"x":7,"y":-1.5865},{"x":8,"y":0.59129},{"x":9,"y":-0.99314},{"x":10,"y":-1.0829},{"x":11,"y":-0.17441},{"x":12,"y":-0.99972},{"x":13,"y":-0.8976},{"x":14,"y":-0.57532},{"x":15,"y":-0.92552},{"x":16,"y":-0.91328},{"x":17,"y":-1.1929},{"x":18,"y":-0.65438},{"x":19,"y":-1.657},{"x":20,"y":-0.025507},{"x":21,"y":-0.77629},{"x":22,"y":-0.7095},{"x":23,"y":0.1119},{"x":24,"y":-1.5459},{"x":25,"y":0.66598},{"x":26,"y":-1.02},{"x":27,"y":-0.87674},{"x":28,"y":-0.15425},{"x":29,"y":-1.1489},{"x":30,"y":-0.82843},{"x":31,"y":-0.64842},{"x":32,"y":-0.82055},{"x":33,"y":-1.0004},{"x":34,"y":-1.0683},{"x":35,"y":-0.93447},{"x":36,"y":-1.2922},{"x":37,"y":-0.35682},{"x":38,"y":-1.3844},{"x":39,"y":0.10704},{"x":40,"y":-0.93704},{"x":41,"y":-0.5491},{"x":42,"y":-0.474},{"x":43,"y":-1.585},{"x":44,"y":-0.22502},{"x":45,"y":-0.97659},{"x":46,"y":-0.33674},{"x":47,"y":-0.60478},{"x":48,"y":-1.2616},{"x":49,"y":-0.52966}]},{"name":"7f03be47-136a-450e-b2be-19d891d2b057","values":[{"x":0,"y":-0.7762485094834298},{"x":1,"y":-0.6748370963030034},{"x":2,"y":-1.7058273929151164},{"x":3,"y":-0.050466594746159485},{"x":4,"y":-0.7755159431653998},{"x":5,"y":-0.7891177507558368},{"x":6,"y":0.04755839805937502},{"x":7,"y":-1.591889606495356},{"x":8,"y":0.431931724948218},{"x":9,"y":-1.0997535743455151},{"x":10,"y":-0.7792837287971803},{"x":11,"y":-0.08358655840536286},{"x":12,"y":-1.614304572536013},{"x":13,"y":-0.3675207278455841},{"x":14,"y":-0.6998160775335086},{"x":15,"y":-0.5703177444418763},{"x":16,"y":-1.125625631733454},{"x":17,"y":-1.221481522436758},{"x":18,"y":-0.7966144757099916},{"x":19,"y":-1.614972779344428},{"x":20,"y":-0.0035628342994057337},{"x":21,"y":-0.6494191707139744},{"x":22,"y":-0.9339536847612101},{"x":23,"y":0.139379895035778},{"x":24,"y":-1.5379121280738701},{"x":25,"y":0.5000730304915373},{"x":26,"y":-1.0038415414621382},{"x":27,"y":-0.5342487194066186},{"x":28,"y":-0.41866402663838653},{"x":29,"y":-1.5086602745604956},{"x":30,"y":-0.3009207934109239},{"x":31,"y":-1.0148594621169562},{"x":32,"y":-0.43167251040313154},{"x":33,"y":-0.8958971733491156},{"x":34,"y":-1.1756644280444644},{"x":35,"y":-0.9499850012245052},{"x":36,"y":-1.3376897383744646},{"x":37,"y":-0.43033522608923835},{"x":38,"y":-1.5141032794247729},{"x":39,"y":-0.10859297046424359},{"x":40,"y":-0.8683544593022704},{"x":41,"y":-0.5946948823966152},{"x":42,"y":-0.7554256380648477},{"x":43,"y":-1.6296589574230467},{"x":44,"y":-0.2698212492202462},{"x":45,"y":-1.3397192082615696},{"x":46,"y":0.028165604793616218},{"x":47,"y":-0.9177871242946081},{"x":48,"y":-0.5743534996619444},{"x":49,"y":-0.9970731323220042}]}],"marks":[{"type":"line","from":{"data":"63c708e6-946d-4d57-8874-b6722b2ff366"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"blue"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}},{"type":"line","from":{"data":"7f03be47-136a-450e-b2be-19d891d2b057"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"red"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}}]},"value":"#gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"63c708e6-946d-4d57-8874-b6722b2ff366\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"63c708e6-946d-4d57-8874-b6722b2ff366\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"63c708e6-946d-4d57-8874-b6722b2ff366\", :values ({:x 0, :y -0.7683} {:x 1, :y -0.37145} {:x 2, :y -1.7936} {:x 3, :y -0.067909} {:x 4, :y -0.72467} {:x 5, :y -0.7811} {:x 6, :y 0.010419} {:x 7, :y -1.5865} {:x 8, :y 0.59129} {:x 9, :y -0.99314} {:x 10, :y -1.0829} {:x 11, :y -0.17441} {:x 12, :y -0.99972} {:x 13, :y -0.8976} {:x 14, :y -0.57532} {:x 15, :y -0.92552} {:x 16, :y -0.91328} {:x 17, :y -1.1929} {:x 18, :y -0.65438} {:x 19, :y -1.657} {:x 20, :y -0.025507} {:x 21, :y -0.77629} {:x 22, :y -0.7095} {:x 23, :y 0.1119} {:x 24, :y -1.5459} {:x 25, :y 0.66598} {:x 26, :y -1.02} {:x 27, :y -0.87674} {:x 28, :y -0.15425} {:x 29, :y -1.1489} {:x 30, :y -0.82843} {:x 31, :y -0.64842} {:x 32, :y -0.82055} {:x 33, :y -1.0004} {:x 34, :y -1.0683} {:x 35, :y -0.93447} {:x 36, :y -1.2922} {:x 37, :y -0.35682} {:x 38, :y -1.3844} {:x 39, :y 0.10704} {:x 40, :y -0.93704} {:x 41, :y -0.5491} {:x 42, :y -0.474} {:x 43, :y -1.585} {:x 44, :y -0.22502} {:x 45, :y -0.97659} {:x 46, :y -0.33674} {:x 47, :y -0.60478} {:x 48, :y -1.2616} {:x 49, :y -0.52966})} {:name \"7f03be47-136a-450e-b2be-19d891d2b057\", :values ({:x 0, :y -0.7762485094834298} {:x 1, :y -0.6748370963030034} {:x 2, :y -1.7058273929151164} {:x 3, :y -0.050466594746159485} {:x 4, :y -0.7755159431653998} {:x 5, :y -0.7891177507558368} {:x 6, :y 0.04755839805937502} {:x 7, :y -1.591889606495356} {:x 8, :y 0.431931724948218} {:x 9, :y -1.0997535743455151} {:x 10, :y -0.7792837287971803} {:x 11, :y -0.08358655840536286} {:x 12, :y -1.614304572536013} {:x 13, :y -0.3675207278455841} {:x 14, :y -0.6998160775335086} {:x 15, :y -0.5703177444418763} {:x 16, :y -1.125625631733454} {:x 17, :y -1.221481522436758} {:x 18, :y -0.7966144757099916} {:x 19, :y -1.614972779344428} {:x 20, :y -0.0035628342994057337} {:x 21, :y -0.6494191707139744} {:x 22, :y -0.9339536847612101} {:x 23, :y 0.139379895035778} {:x 24, :y -1.5379121280738701} {:x 25, :y 0.5000730304915373} {:x 26, :y -1.0038415414621382} {:x 27, :y -0.5342487194066186} {:x 28, :y -0.41866402663838653} {:x 29, :y -1.5086602745604956} {:x 30, :y -0.3009207934109239} {:x 31, :y -1.0148594621169562} {:x 32, :y -0.43167251040313154} {:x 33, :y -0.8958971733491156} {:x 34, :y -1.1756644280444644} {:x 35, :y -0.9499850012245052} {:x 36, :y -1.3376897383744646} {:x 37, :y -0.43033522608923835} {:x 38, :y -1.5141032794247729} {:x 39, :y -0.10859297046424359} {:x 40, :y -0.8683544593022704} {:x 41, :y -0.5946948823966152} {:x 42, :y -0.7554256380648477} {:x 43, :y -1.6296589574230467} {:x 44, :y -0.2698212492202462} {:x 45, :y -1.3397192082615696} {:x 46, :y 0.028165604793616218} {:x 47, :y -0.9177871242946081} {:x 48, :y -0.5743534996619444} {:x 49, :y -0.9970731323220042})}), :marks ({:type \"line\", :from {:data \"63c708e6-946d-4d57-8874-b6722b2ff366\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"7f03be47-136a-450e-b2be-19d891d2b057\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}}"},{"type":"vega","content":{"width":250,"height":154.51174926757812,"padding":{"top":10,"left":55,"bottom":40,"right":10},"scales":[{"name":"x","type":"linear","range":"width","zero":false,"domain":{"data":"12c3e3f3-0d1e-4e39-8316-e5c8a57070ca","field":"data.x"}},{"name":"y","type":"linear","range":"height","nice":true,"zero":false,"domain":{"data":"12c3e3f3-0d1e-4e39-8316-e5c8a57070ca","field":"data.y"}}],"axes":[{"type":"x","scale":"x","title":"Time step","titleOffset":30},{"type":"y","scale":"y"}],"data":[{"name":"12c3e3f3-0d1e-4e39-8316-e5c8a57070ca","values":[{"x":0,"y":-0.23752},{"x":1,"y":0.98831},{"x":2,"y":0.69057},{"x":3,"y":0.88147},{"x":4,"y":-0.2074},{"x":5,"y":-0.57156},{"x":6,"y":0.84455},{"x":7,"y":0.48982},{"x":8,"y":-0.2632},{"x":9,"y":-0.76504},{"x":10,"y":-0.90355},{"x":11,"y":0.030318},{"x":12,"y":0.99796},{"x":13,"y":0.40324},{"x":14,"y":0.23665},{"x":15,"y":0.52027},{"x":16,"y":0.79671},{"x":17,"y":0.93087},{"x":18,"y":0.99681},{"x":19,"y":0.78457},{"x":20,"y":0.99837},{"x":21,"y":-0.14975},{"x":22,"y":-0.76437},{"x":23,"y":0.7783},{"x":24,"y":0.50557},{"x":25,"y":-0.42942},{"x":26,"y":-0.58685},{"x":27,"y":-0.84038},{"x":28,"y":0.24341},{"x":29,"y":0.99984},{"x":30,"y":0.48954},{"x":31,"y":0.18986},{"x":32,"y":0.48798},{"x":33,"y":0.80838},{"x":34,"y":0.96206},{"x":35,"y":0.99978},{"x":36,"y":0.97508},{"x":37,"y":0.97114},{"x":38,"y":0.33953},{"x":39,"y":0.10228},{"x":40,"y":-0.36249},{"x":41,"y":-0.30036},{"x":42,"y":0.88944},{"x":43,"y":0.62605},{"x":44,"y":0.83797},{"x":45,"y":0.012639},{"x":46,"y":-0.33613},{"x":47,"y":0.69772},{"x":48,"y":0.54889},{"x":49,"y":0.78432}]},{"name":"bbbeda71-dc6b-4194-8ee4-43dc89b34d48","values":[{"x":0,"y":0.02636840598654447},{"x":1,"y":0.9671116038360211},{"x":2,"y":0.8716024902496547},{"x":3,"y":0.9504242821101277},{"x":4,"y":-0.1604606197739061},{"x":5,"y":-0.7029732766574083},{"x":6,"y":0.8189784569470407},{"x":7,"y":0.5126662761097954},{"x":8,"y":-0.2916788339311217},{"x":9,"y":-0.4895880706147808},{"x":10,"y":-0.6399126434122681},{"x":11,"y":0.6249405961929843},{"x":12,"y":0.7801774459615055},{"x":13,"y":0.5194720761282753},{"x":14,"y":-0.2937747553284505},{"x":15,"y":0.31288094684670514},{"x":16,"y":0.9314082268375062},{"x":17,"y":0.9520428780508164},{"x":18,"y":0.9943393221058118},{"x":19,"y":0.9015471319470735},{"x":20,"y":0.931531233509654},{"x":21,"y":-0.24077943374478034},{"x":22,"y":-0.6515949385523324},{"x":23,"y":0.786701764124919},{"x":24,"y":0.5070632983713366},{"x":25,"y":-0.4813551658599975},{"x":26,"y":-0.17575402189296227},{"x":27,"y":-0.4104099833161618},{"x":28,"y":0.8152433363158724},{"x":29,"y":0.5210117678235281},{"x":30,"y":0.6043859570591666},{"x":31,"y":0.07358405188981251},{"x":32,"y":-3.551481974257281E-4},{"x":33,"y":0.7099342662667935},{"x":34,"y":0.6981853472761372},{"x":35,"y":0.9458801520298787},{"x":36,"y":0.9631611844236628},{"x":37,"y":0.9612269599102076},{"x":38,"y":0.4395392031765674},{"x":39,"y":0.34612157207138045},{"x":40,"y":-0.2455184737013482},{"x":41,"y":-0.03593680200133874},{"x":42,"y":0.9793052484540875},{"x":43,"y":0.7824717775390854},{"x":44,"y":1.011923195821054},{"x":45,"y":0.26988635735669175},{"x":46,"y":-0.2934883360831475},{"x":47,"y":0.17705902474888938},{"x":48,"y":0.2132589287448789},{"x":49,"y":0.6891729608705063}]}],"marks":[{"type":"line","from":{"data":"12c3e3f3-0d1e-4e39-8316-e5c8a57070ca"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"blue"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}},{"type":"line","from":{"data":"bbbeda71-dc6b-4194-8ee4-43dc89b34d48"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"red"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}}]},"value":"#gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"12c3e3f3-0d1e-4e39-8316-e5c8a57070ca\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"12c3e3f3-0d1e-4e39-8316-e5c8a57070ca\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"12c3e3f3-0d1e-4e39-8316-e5c8a57070ca\", :values ({:x 0, :y -0.23752} {:x 1, :y 0.98831} {:x 2, :y 0.69057} {:x 3, :y 0.88147} {:x 4, :y -0.2074} {:x 5, :y -0.57156} {:x 6, :y 0.84455} {:x 7, :y 0.48982} {:x 8, :y -0.2632} {:x 9, :y -0.76504} {:x 10, :y -0.90355} {:x 11, :y 0.030318} {:x 12, :y 0.99796} {:x 13, :y 0.40324} {:x 14, :y 0.23665} {:x 15, :y 0.52027} {:x 16, :y 0.79671} {:x 17, :y 0.93087} {:x 18, :y 0.99681} {:x 19, :y 0.78457} {:x 20, :y 0.99837} {:x 21, :y -0.14975} {:x 22, :y -0.76437} {:x 23, :y 0.7783} {:x 24, :y 0.50557} {:x 25, :y -0.42942} {:x 26, :y -0.58685} {:x 27, :y -0.84038} {:x 28, :y 0.24341} {:x 29, :y 0.99984} {:x 30, :y 0.48954} {:x 31, :y 0.18986} {:x 32, :y 0.48798} {:x 33, :y 0.80838} {:x 34, :y 0.96206} {:x 35, :y 0.99978} {:x 36, :y 0.97508} {:x 37, :y 0.97114} {:x 38, :y 0.33953} {:x 39, :y 0.10228} {:x 40, :y -0.36249} {:x 41, :y -0.30036} {:x 42, :y 0.88944} {:x 43, :y 0.62605} {:x 44, :y 0.83797} {:x 45, :y 0.012639} {:x 46, :y -0.33613} {:x 47, :y 0.69772} {:x 48, :y 0.54889} {:x 49, :y 0.78432})} {:name \"bbbeda71-dc6b-4194-8ee4-43dc89b34d48\", :values ({:x 0, :y 0.02636840598654447} {:x 1, :y 0.9671116038360211} {:x 2, :y 0.8716024902496547} {:x 3, :y 0.9504242821101277} {:x 4, :y -0.1604606197739061} {:x 5, :y -0.7029732766574083} {:x 6, :y 0.8189784569470407} {:x 7, :y 0.5126662761097954} {:x 8, :y -0.2916788339311217} {:x 9, :y -0.4895880706147808} {:x 10, :y -0.6399126434122681} {:x 11, :y 0.6249405961929843} {:x 12, :y 0.7801774459615055} {:x 13, :y 0.5194720761282753} {:x 14, :y -0.2937747553284505} {:x 15, :y 0.31288094684670514} {:x 16, :y 0.9314082268375062} {:x 17, :y 0.9520428780508164} {:x 18, :y 0.9943393221058118} {:x 19, :y 0.9015471319470735} {:x 20, :y 0.931531233509654} {:x 21, :y -0.24077943374478034} {:x 22, :y -0.6515949385523324} {:x 23, :y 0.786701764124919} {:x 24, :y 0.5070632983713366} {:x 25, :y -0.4813551658599975} {:x 26, :y -0.17575402189296227} {:x 27, :y -0.4104099833161618} {:x 28, :y 0.8152433363158724} {:x 29, :y 0.5210117678235281} {:x 30, :y 0.6043859570591666} {:x 31, :y 0.07358405188981251} {:x 32, :y -3.551481974257281E-4} {:x 33, :y 0.7099342662667935} {:x 34, :y 0.6981853472761372} {:x 35, :y 0.9458801520298787} {:x 36, :y 0.9631611844236628} {:x 37, :y 0.9612269599102076} {:x 38, :y 0.4395392031765674} {:x 39, :y 0.34612157207138045} {:x 40, :y -0.2455184737013482} {:x 41, :y -0.03593680200133874} {:x 42, :y 0.9793052484540875} {:x 43, :y 0.7824717775390854} {:x 44, :y 1.011923195821054} {:x 45, :y 0.26988635735669175} {:x 46, :y -0.2934883360831475} {:x 47, :y 0.17705902474888938} {:x 48, :y 0.2132589287448789} {:x 49, :y 0.6891729608705063})}), :marks ({:type \"line\", :from {:data \"12c3e3f3-0d1e-4e39-8316-e5c8a57070ca\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"bbbeda71-dc6b-4194-8ee4-43dc89b34d48\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}}"}],"value":"[#gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"32c07c23-a435-45b6-9bf9-abec1ba19a41\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"32c07c23-a435-45b6-9bf9-abec1ba19a41\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"32c07c23-a435-45b6-9bf9-abec1ba19a41\", :values ({:x 0, :y 1.4177} {:x 1, :y 0.76228} {:x 2, :y 1.079} {:x 3, :y -0.20891} {:x 4, :y -0.6084} {:x 5, :y 1.0057} {:x 6, :y 0.51189} {:x 7, :y -0.26634} {:x 8, :y -0.8711} {:x 9, :y -1.128} {:x 10, :y 0.030323} {:x 11, :y 1.5069} {:x 12, :y 0.41506} {:x 13, :y 0.23892} {:x 14, :y 0.54717} {:x 15, :y 0.92183} {:x 16, :y 1.1968} {:x 17, :y 1.6507} {:x 18, :y 0.90201} {:x 19, :y 1.6279} {:x 20, :y -0.15032} {:x 21, :y -0.87007} {:x 22, :y 0.89195} {:x 23, :y 0.53004} {:x 24, :y -0.44385} {:x 25, :y -0.62716} {:x 26, :y -0.99799} {:x 27, :y 0.24588} {:x 28, :y 1.5889} {:x 29, :y 0.51156} {:x 30, :y 0.19102} {:x 31, :y 0.50977} {:x 32, :y 0.9414} {:x 33, :y 1.2945} {:x 34, :y 1.5499} {:x 35, :y 1.3471} {:x 36, :y 1.8116} {:x 37, :y 0.34641} {:x 38, :y 0.10246} {:x 39, :y -0.37094} {:x 40, :y -0.30507} {:x 41, :y 1.0961} {:x 42, :y 0.67648} {:x 43, :y 0.99356} {:x 44, :y 0.012639} {:x 45, :y -0.3428} {:x 46, :y 0.77221} {:x 47, :y 0.58104} {:x 48, :y 0.9016} {:x 49, :y 0.58413})} {:name \"573e36c4-f2f4-4afb-9bcb-6f4d5072396c\", :values ({:x 0, :y 1.3402538531627508} {:x 1, :y 1.0322002619864143} {:x 2, :y 1.8014341933167113} {:x 3, :y -0.16000458313615792} {:x 4, :y -0.7744483571217643} {:x 5, :y 0.9363745319764447} {:x 6, :y 0.5254895921062439} {:x 7, :y -0.29915283938028475} {:x 8, :y -0.5038928058727293} {:x 9, :y -0.6955922618142655} {:x 10, :y 0.6898091803077965} {:x 11, :y 0.9016269297091006} {:x 12, :y 0.557289695078088} {:x 13, :y -0.2984692001144609} {:x 14, :y 0.32091215648940197} {:x 15, :y 1.1914191725037722} {:x 16, :y 1.2295436136266658} {:x 17, :y 1.6593034716783146} {:x 18, :y 1.110223671772022} {:x 19, :y 1.9275870897723504} {:x 20, :y -0.25857193351522606} {:x 21, :y -0.7170811459608197} {:x 22, :y 0.9169626198614169} {:x 23, :y 0.5224845031148656} {:x 24, :y -0.49460789861013116} {:x 25, :y -0.1857460198076708} {:x 26, :y -0.4231543651316923} {:x 27, :y 0.9685359407715033} {:x 28, :y 0.5723638113647693} {:x 29, :y 0.6439790481989783} {:x 30, :y 0.07648367501862453} {:x 31, :y -0.009815628541591737} {:x 32, :y 0.7953490708998123} {:x 33, :y 0.7743124109915819} {:x 34, :y 1.207594294421123} {:x 35, :y 1.352987814356697} {:x 36, :y 1.8590954334985326} {:x 37, :y 0.4543099495878354} {:x 38, :y 0.35608643974543747} {:x 39, :y -0.24257402352852445} {:x 40, :y -0.0386439876811051} {:x 41, :y 1.2391330793554265} {:x 42, :y 0.903430610994317} {:x 43, :y 1.6197532790143159} {:x 44, :y 0.2611887771032473} {:x 45, :y -0.29700438654409667} {:x 46, :y 0.1841301329648706} {:x 47, :y 0.2128680631461865} {:x 48, :y 0.7816724284615412} {:x 49, :y 1.007302031506389})}), :marks ({:type \"line\", :from {:data \"32c07c23-a435-45b6-9bf9-abec1ba19a41\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"573e36c4-f2f4-4afb-9bcb-6f4d5072396c\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}} #gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"63c708e6-946d-4d57-8874-b6722b2ff366\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"63c708e6-946d-4d57-8874-b6722b2ff366\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"63c708e6-946d-4d57-8874-b6722b2ff366\", :values ({:x 0, :y -0.7683} {:x 1, :y -0.37145} {:x 2, :y -1.7936} {:x 3, :y -0.067909} {:x 4, :y -0.72467} {:x 5, :y -0.7811} {:x 6, :y 0.010419} {:x 7, :y -1.5865} {:x 8, :y 0.59129} {:x 9, :y -0.99314} {:x 10, :y -1.0829} {:x 11, :y -0.17441} {:x 12, :y -0.99972} {:x 13, :y -0.8976} {:x 14, :y -0.57532} {:x 15, :y -0.92552} {:x 16, :y -0.91328} {:x 17, :y -1.1929} {:x 18, :y -0.65438} {:x 19, :y -1.657} {:x 20, :y -0.025507} {:x 21, :y -0.77629} {:x 22, :y -0.7095} {:x 23, :y 0.1119} {:x 24, :y -1.5459} {:x 25, :y 0.66598} {:x 26, :y -1.02} {:x 27, :y -0.87674} {:x 28, :y -0.15425} {:x 29, :y -1.1489} {:x 30, :y -0.82843} {:x 31, :y -0.64842} {:x 32, :y -0.82055} {:x 33, :y -1.0004} {:x 34, :y -1.0683} {:x 35, :y -0.93447} {:x 36, :y -1.2922} {:x 37, :y -0.35682} {:x 38, :y -1.3844} {:x 39, :y 0.10704} {:x 40, :y -0.93704} {:x 41, :y -0.5491} {:x 42, :y -0.474} {:x 43, :y -1.585} {:x 44, :y -0.22502} {:x 45, :y -0.97659} {:x 46, :y -0.33674} {:x 47, :y -0.60478} {:x 48, :y -1.2616} {:x 49, :y -0.52966})} {:name \"7f03be47-136a-450e-b2be-19d891d2b057\", :values ({:x 0, :y -0.7762485094834298} {:x 1, :y -0.6748370963030034} {:x 2, :y -1.7058273929151164} {:x 3, :y -0.050466594746159485} {:x 4, :y -0.7755159431653998} {:x 5, :y -0.7891177507558368} {:x 6, :y 0.04755839805937502} {:x 7, :y -1.591889606495356} {:x 8, :y 0.431931724948218} {:x 9, :y -1.0997535743455151} {:x 10, :y -0.7792837287971803} {:x 11, :y -0.08358655840536286} {:x 12, :y -1.614304572536013} {:x 13, :y -0.3675207278455841} {:x 14, :y -0.6998160775335086} {:x 15, :y -0.5703177444418763} {:x 16, :y -1.125625631733454} {:x 17, :y -1.221481522436758} {:x 18, :y -0.7966144757099916} {:x 19, :y -1.614972779344428} {:x 20, :y -0.0035628342994057337} {:x 21, :y -0.6494191707139744} {:x 22, :y -0.9339536847612101} {:x 23, :y 0.139379895035778} {:x 24, :y -1.5379121280738701} {:x 25, :y 0.5000730304915373} {:x 26, :y -1.0038415414621382} {:x 27, :y -0.5342487194066186} {:x 28, :y -0.41866402663838653} {:x 29, :y -1.5086602745604956} {:x 30, :y -0.3009207934109239} {:x 31, :y -1.0148594621169562} {:x 32, :y -0.43167251040313154} {:x 33, :y -0.8958971733491156} {:x 34, :y -1.1756644280444644} {:x 35, :y -0.9499850012245052} {:x 36, :y -1.3376897383744646} {:x 37, :y -0.43033522608923835} {:x 38, :y -1.5141032794247729} {:x 39, :y -0.10859297046424359} {:x 40, :y -0.8683544593022704} {:x 41, :y -0.5946948823966152} {:x 42, :y -0.7554256380648477} {:x 43, :y -1.6296589574230467} {:x 44, :y -0.2698212492202462} {:x 45, :y -1.3397192082615696} {:x 46, :y 0.028165604793616218} {:x 47, :y -0.9177871242946081} {:x 48, :y -0.5743534996619444} {:x 49, :y -0.9970731323220042})}), :marks ({:type \"line\", :from {:data \"63c708e6-946d-4d57-8874-b6722b2ff366\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"7f03be47-136a-450e-b2be-19d891d2b057\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}} #gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"12c3e3f3-0d1e-4e39-8316-e5c8a57070ca\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"12c3e3f3-0d1e-4e39-8316-e5c8a57070ca\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"12c3e3f3-0d1e-4e39-8316-e5c8a57070ca\", :values ({:x 0, :y -0.23752} {:x 1, :y 0.98831} {:x 2, :y 0.69057} {:x 3, :y 0.88147} {:x 4, :y -0.2074} {:x 5, :y -0.57156} {:x 6, :y 0.84455} {:x 7, :y 0.48982} {:x 8, :y -0.2632} {:x 9, :y -0.76504} {:x 10, :y -0.90355} {:x 11, :y 0.030318} {:x 12, :y 0.99796} {:x 13, :y 0.40324} {:x 14, :y 0.23665} {:x 15, :y 0.52027} {:x 16, :y 0.79671} {:x 17, :y 0.93087} {:x 18, :y 0.99681} {:x 19, :y 0.78457} {:x 20, :y 0.99837} {:x 21, :y -0.14975} {:x 22, :y -0.76437} {:x 23, :y 0.7783} {:x 24, :y 0.50557} {:x 25, :y -0.42942} {:x 26, :y -0.58685} {:x 27, :y -0.84038} {:x 28, :y 0.24341} {:x 29, :y 0.99984} {:x 30, :y 0.48954} {:x 31, :y 0.18986} {:x 32, :y 0.48798} {:x 33, :y 0.80838} {:x 34, :y 0.96206} {:x 35, :y 0.99978} {:x 36, :y 0.97508} {:x 37, :y 0.97114} {:x 38, :y 0.33953} {:x 39, :y 0.10228} {:x 40, :y -0.36249} {:x 41, :y -0.30036} {:x 42, :y 0.88944} {:x 43, :y 0.62605} {:x 44, :y 0.83797} {:x 45, :y 0.012639} {:x 46, :y -0.33613} {:x 47, :y 0.69772} {:x 48, :y 0.54889} {:x 49, :y 0.78432})} {:name \"bbbeda71-dc6b-4194-8ee4-43dc89b34d48\", :values ({:x 0, :y 0.02636840598654447} {:x 1, :y 0.9671116038360211} {:x 2, :y 0.8716024902496547} {:x 3, :y 0.9504242821101277} {:x 4, :y -0.1604606197739061} {:x 5, :y -0.7029732766574083} {:x 6, :y 0.8189784569470407} {:x 7, :y 0.5126662761097954} {:x 8, :y -0.2916788339311217} {:x 9, :y -0.4895880706147808} {:x 10, :y -0.6399126434122681} {:x 11, :y 0.6249405961929843} {:x 12, :y 0.7801774459615055} {:x 13, :y 0.5194720761282753} {:x 14, :y -0.2937747553284505} {:x 15, :y 0.31288094684670514} {:x 16, :y 0.9314082268375062} {:x 17, :y 0.9520428780508164} {:x 18, :y 0.9943393221058118} {:x 19, :y 0.9015471319470735} {:x 20, :y 0.931531233509654} {:x 21, :y -0.24077943374478034} {:x 22, :y -0.6515949385523324} {:x 23, :y 0.786701764124919} {:x 24, :y 0.5070632983713366} {:x 25, :y -0.4813551658599975} {:x 26, :y -0.17575402189296227} {:x 27, :y -0.4104099833161618} {:x 28, :y 0.8152433363158724} {:x 29, :y 0.5210117678235281} {:x 30, :y 0.6043859570591666} {:x 31, :y 0.07358405188981251} {:x 32, :y -3.551481974257281E-4} {:x 33, :y 0.7099342662667935} {:x 34, :y 0.6981853472761372} {:x 35, :y 0.9458801520298787} {:x 36, :y 0.9631611844236628} {:x 37, :y 0.9612269599102076} {:x 38, :y 0.4395392031765674} {:x 39, :y 0.34612157207138045} {:x 40, :y -0.2455184737013482} {:x 41, :y -0.03593680200133874} {:x 42, :y 0.9793052484540875} {:x 43, :y 0.7824717775390854} {:x 44, :y 1.011923195821054} {:x 45, :y 0.26988635735669175} {:x 46, :y -0.2934883360831475} {:x 47, :y 0.17705902474888938} {:x 48, :y 0.2132589287448789} {:x 49, :y 0.6891729608705063})}), :marks ({:type \"line\", :from {:data \"12c3e3f3-0d1e-4e39-8316-e5c8a57070ca\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"bbbeda71-dc6b-4194-8ee4-43dc89b34d48\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}}]"}
;; <=

;; **
;;; Ground truth (blue) and inferred (red) trajectories. Left to right: dimensions 1 to 3. After **100 iterations of BOPP**.
;; **

;; @@
(mapv 
  #(plot/compose
     (plot/list-plot %1
                     :joined true 
                     :color (first plot-colors)
                     :plot-size 250
                     :x-title "Time step")
     (plot/list-plot %2
                     :joined true 
                     :color (second plot-colors)))
  ground-truth-trajectory
  (nth inferred-trajectories 49))
;; @@
;; =>
;;; {"type":"list-like","open":"<span class='clj-vector'>[</span>","close":"<span class='clj-vector'>]</span>","separator":" ","items":[{"type":"vega","content":{"width":250,"height":154.51174926757812,"padding":{"top":10,"left":55,"bottom":40,"right":10},"scales":[{"name":"x","type":"linear","range":"width","zero":false,"domain":{"data":"f99f98ff-dc50-4e8a-ae17-b6b8afc1bce3","field":"data.x"}},{"name":"y","type":"linear","range":"height","nice":true,"zero":false,"domain":{"data":"f99f98ff-dc50-4e8a-ae17-b6b8afc1bce3","field":"data.y"}}],"axes":[{"type":"x","scale":"x","title":"Time step","titleOffset":30},{"type":"y","scale":"y"}],"data":[{"name":"f99f98ff-dc50-4e8a-ae17-b6b8afc1bce3","values":[{"x":0,"y":1.4177},{"x":1,"y":0.76228},{"x":2,"y":1.079},{"x":3,"y":-0.20891},{"x":4,"y":-0.6084},{"x":5,"y":1.0057},{"x":6,"y":0.51189},{"x":7,"y":-0.26634},{"x":8,"y":-0.8711},{"x":9,"y":-1.128},{"x":10,"y":0.030323},{"x":11,"y":1.5069},{"x":12,"y":0.41506},{"x":13,"y":0.23892},{"x":14,"y":0.54717},{"x":15,"y":0.92183},{"x":16,"y":1.1968},{"x":17,"y":1.6507},{"x":18,"y":0.90201},{"x":19,"y":1.6279},{"x":20,"y":-0.15032},{"x":21,"y":-0.87007},{"x":22,"y":0.89195},{"x":23,"y":0.53004},{"x":24,"y":-0.44385},{"x":25,"y":-0.62716},{"x":26,"y":-0.99799},{"x":27,"y":0.24588},{"x":28,"y":1.5889},{"x":29,"y":0.51156},{"x":30,"y":0.19102},{"x":31,"y":0.50977},{"x":32,"y":0.9414},{"x":33,"y":1.2945},{"x":34,"y":1.5499},{"x":35,"y":1.3471},{"x":36,"y":1.8116},{"x":37,"y":0.34641},{"x":38,"y":0.10246},{"x":39,"y":-0.37094},{"x":40,"y":-0.30507},{"x":41,"y":1.0961},{"x":42,"y":0.67648},{"x":43,"y":0.99356},{"x":44,"y":0.012639},{"x":45,"y":-0.3428},{"x":46,"y":0.77221},{"x":47,"y":0.58104},{"x":48,"y":0.9016},{"x":49,"y":0.58413}]},{"name":"5b09fd8d-b39d-44c8-9140-4d91ad8318c1","values":[{"x":0,"y":1.47789421139643},{"x":1,"y":0.7485151024684631},{"x":2,"y":1.1256099629495988},{"x":3,"y":-0.21853907378491716},{"x":4,"y":-0.7269873612328437},{"x":5,"y":0.9243906448806127},{"x":6,"y":0.5079399323915975},{"x":7,"y":-0.4643609909305077},{"x":8,"y":-0.6938857135381984},{"x":9,"y":-1.0751172609556205},{"x":10,"y":0.11127752883728341},{"x":11,"y":1.5650434986725124},{"x":12,"y":0.5098879052689004},{"x":13,"y":0.2700731527855736},{"x":14,"y":0.4865271103826882},{"x":15,"y":0.891685387701729},{"x":16,"y":1.183537517436241},{"x":17,"y":1.6181818165324042},{"x":18,"y":0.9212837951635541},{"x":19,"y":1.6425641829230153},{"x":20,"y":-0.1637719679958235},{"x":21,"y":-0.8777910035632345},{"x":22,"y":0.886981748947608},{"x":23,"y":0.5247962932184906},{"x":24,"y":-0.508562527666783},{"x":25,"y":-0.5699526124191071},{"x":26,"y":-0.9362095719068727},{"x":27,"y":0.31227804696536793},{"x":28,"y":1.501237006245938},{"x":29,"y":0.6093597664377434},{"x":30,"y":0.18565776324740893},{"x":31,"y":0.44712944601659305},{"x":32,"y":0.9400324458839735},{"x":33,"y":1.2674898358942575},{"x":34,"y":1.5182821258431827},{"x":35,"y":1.3356908825967275},{"x":36,"y":1.7984611996374504},{"x":37,"y":0.3372543931653316},{"x":38,"y":0.14931963354971414},{"x":39,"y":-0.4049218506098873},{"x":40,"y":-0.266311495138262},{"x":41,"y":1.137697659887803},{"x":42,"y":0.727910766774785},{"x":43,"y":1.1380352685066726},{"x":44,"y":0.014122045134247082},{"x":45,"y":-0.5196234742375875},{"x":46,"y":0.6973920033661354},{"x":47,"y":0.576054622226453},{"x":48,"y":0.8254914141780186},{"x":49,"y":0.3288001151069147}]}],"marks":[{"type":"line","from":{"data":"f99f98ff-dc50-4e8a-ae17-b6b8afc1bce3"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"blue"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}},{"type":"line","from":{"data":"5b09fd8d-b39d-44c8-9140-4d91ad8318c1"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"red"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}}]},"value":"#gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"f99f98ff-dc50-4e8a-ae17-b6b8afc1bce3\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"f99f98ff-dc50-4e8a-ae17-b6b8afc1bce3\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"f99f98ff-dc50-4e8a-ae17-b6b8afc1bce3\", :values ({:x 0, :y 1.4177} {:x 1, :y 0.76228} {:x 2, :y 1.079} {:x 3, :y -0.20891} {:x 4, :y -0.6084} {:x 5, :y 1.0057} {:x 6, :y 0.51189} {:x 7, :y -0.26634} {:x 8, :y -0.8711} {:x 9, :y -1.128} {:x 10, :y 0.030323} {:x 11, :y 1.5069} {:x 12, :y 0.41506} {:x 13, :y 0.23892} {:x 14, :y 0.54717} {:x 15, :y 0.92183} {:x 16, :y 1.1968} {:x 17, :y 1.6507} {:x 18, :y 0.90201} {:x 19, :y 1.6279} {:x 20, :y -0.15032} {:x 21, :y -0.87007} {:x 22, :y 0.89195} {:x 23, :y 0.53004} {:x 24, :y -0.44385} {:x 25, :y -0.62716} {:x 26, :y -0.99799} {:x 27, :y 0.24588} {:x 28, :y 1.5889} {:x 29, :y 0.51156} {:x 30, :y 0.19102} {:x 31, :y 0.50977} {:x 32, :y 0.9414} {:x 33, :y 1.2945} {:x 34, :y 1.5499} {:x 35, :y 1.3471} {:x 36, :y 1.8116} {:x 37, :y 0.34641} {:x 38, :y 0.10246} {:x 39, :y -0.37094} {:x 40, :y -0.30507} {:x 41, :y 1.0961} {:x 42, :y 0.67648} {:x 43, :y 0.99356} {:x 44, :y 0.012639} {:x 45, :y -0.3428} {:x 46, :y 0.77221} {:x 47, :y 0.58104} {:x 48, :y 0.9016} {:x 49, :y 0.58413})} {:name \"5b09fd8d-b39d-44c8-9140-4d91ad8318c1\", :values ({:x 0, :y 1.47789421139643} {:x 1, :y 0.7485151024684631} {:x 2, :y 1.1256099629495988} {:x 3, :y -0.21853907378491716} {:x 4, :y -0.7269873612328437} {:x 5, :y 0.9243906448806127} {:x 6, :y 0.5079399323915975} {:x 7, :y -0.4643609909305077} {:x 8, :y -0.6938857135381984} {:x 9, :y -1.0751172609556205} {:x 10, :y 0.11127752883728341} {:x 11, :y 1.5650434986725124} {:x 12, :y 0.5098879052689004} {:x 13, :y 0.2700731527855736} {:x 14, :y 0.4865271103826882} {:x 15, :y 0.891685387701729} {:x 16, :y 1.183537517436241} {:x 17, :y 1.6181818165324042} {:x 18, :y 0.9212837951635541} {:x 19, :y 1.6425641829230153} {:x 20, :y -0.1637719679958235} {:x 21, :y -0.8777910035632345} {:x 22, :y 0.886981748947608} {:x 23, :y 0.5247962932184906} {:x 24, :y -0.508562527666783} {:x 25, :y -0.5699526124191071} {:x 26, :y -0.9362095719068727} {:x 27, :y 0.31227804696536793} {:x 28, :y 1.501237006245938} {:x 29, :y 0.6093597664377434} {:x 30, :y 0.18565776324740893} {:x 31, :y 0.44712944601659305} {:x 32, :y 0.9400324458839735} {:x 33, :y 1.2674898358942575} {:x 34, :y 1.5182821258431827} {:x 35, :y 1.3356908825967275} {:x 36, :y 1.7984611996374504} {:x 37, :y 0.3372543931653316} {:x 38, :y 0.14931963354971414} {:x 39, :y -0.4049218506098873} {:x 40, :y -0.266311495138262} {:x 41, :y 1.137697659887803} {:x 42, :y 0.727910766774785} {:x 43, :y 1.1380352685066726} {:x 44, :y 0.014122045134247082} {:x 45, :y -0.5196234742375875} {:x 46, :y 0.6973920033661354} {:x 47, :y 0.576054622226453} {:x 48, :y 0.8254914141780186} {:x 49, :y 0.3288001151069147})}), :marks ({:type \"line\", :from {:data \"f99f98ff-dc50-4e8a-ae17-b6b8afc1bce3\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"5b09fd8d-b39d-44c8-9140-4d91ad8318c1\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}}"},{"type":"vega","content":{"width":250,"height":154.51174926757812,"padding":{"top":10,"left":55,"bottom":40,"right":10},"scales":[{"name":"x","type":"linear","range":"width","zero":false,"domain":{"data":"583c11d5-22f8-4df8-97dd-1d56c5e16743","field":"data.x"}},{"name":"y","type":"linear","range":"height","nice":true,"zero":false,"domain":{"data":"583c11d5-22f8-4df8-97dd-1d56c5e16743","field":"data.y"}}],"axes":[{"type":"x","scale":"x","title":"Time step","titleOffset":30},{"type":"y","scale":"y"}],"data":[{"name":"583c11d5-22f8-4df8-97dd-1d56c5e16743","values":[{"x":0,"y":-0.7683},{"x":1,"y":-0.37145},{"x":2,"y":-1.7936},{"x":3,"y":-0.067909},{"x":4,"y":-0.72467},{"x":5,"y":-0.7811},{"x":6,"y":0.010419},{"x":7,"y":-1.5865},{"x":8,"y":0.59129},{"x":9,"y":-0.99314},{"x":10,"y":-1.0829},{"x":11,"y":-0.17441},{"x":12,"y":-0.99972},{"x":13,"y":-0.8976},{"x":14,"y":-0.57532},{"x":15,"y":-0.92552},{"x":16,"y":-0.91328},{"x":17,"y":-1.1929},{"x":18,"y":-0.65438},{"x":19,"y":-1.657},{"x":20,"y":-0.025507},{"x":21,"y":-0.77629},{"x":22,"y":-0.7095},{"x":23,"y":0.1119},{"x":24,"y":-1.5459},{"x":25,"y":0.66598},{"x":26,"y":-1.02},{"x":27,"y":-0.87674},{"x":28,"y":-0.15425},{"x":29,"y":-1.1489},{"x":30,"y":-0.82843},{"x":31,"y":-0.64842},{"x":32,"y":-0.82055},{"x":33,"y":-1.0004},{"x":34,"y":-1.0683},{"x":35,"y":-0.93447},{"x":36,"y":-1.2922},{"x":37,"y":-0.35682},{"x":38,"y":-1.3844},{"x":39,"y":0.10704},{"x":40,"y":-0.93704},{"x":41,"y":-0.5491},{"x":42,"y":-0.474},{"x":43,"y":-1.585},{"x":44,"y":-0.22502},{"x":45,"y":-0.97659},{"x":46,"y":-0.33674},{"x":47,"y":-0.60478},{"x":48,"y":-1.2616},{"x":49,"y":-0.52966}]},{"name":"1c8565e7-5c1f-4a3a-b581-54971972ec90","values":[{"x":0,"y":-0.6561510531194803},{"x":1,"y":-0.42849807387447136},{"x":2,"y":-1.748285969489911},{"x":3,"y":-0.02665405159870647},{"x":4,"y":-0.7205401004571231},{"x":5,"y":-0.7782588434464333},{"x":6,"y":0.1062867823659176},{"x":7,"y":-1.5453853065686403},{"x":8,"y":0.70122101142841},{"x":9,"y":-0.9862039366024559},{"x":10,"y":-0.9735605431468249},{"x":11,"y":-0.18177468462366195},{"x":12,"y":-1.0782262963383853},{"x":13,"y":-0.8794486023600043},{"x":14,"y":-0.6328507740501725},{"x":15,"y":-0.8539690910314582},{"x":16,"y":-0.9031333403235883},{"x":17,"y":-1.1723692942767856},{"x":18,"y":-0.6901353453659438},{"x":19,"y":-1.6055778497237165},{"x":20,"y":-0.02093379587293824},{"x":21,"y":-0.7630999385911313},{"x":22,"y":-0.6946028370513873},{"x":23,"y":0.1298750459960506},{"x":24,"y":-1.504149684160266},{"x":25,"y":0.6932713282146231},{"x":26,"y":-1.0117810104295706},{"x":27,"y":-0.8008590026154816},{"x":28,"y":-0.15036195989404377},{"x":29,"y":-1.2252812466442942},{"x":30,"y":-0.8090341915281736},{"x":31,"y":-0.6848263617135942},{"x":32,"y":-0.7407316346822757},{"x":33,"y":-0.9961792431081292},{"x":34,"y":-1.0676967658287946},{"x":35,"y":-0.9245987316548181},{"x":36,"y":-1.2666592803119894},{"x":37,"y":-0.37849975854020196},{"x":38,"y":-1.356994686877228},{"x":39,"y":0.08629557617885042},{"x":40,"y":-0.9094804325309619},{"x":41,"y":-0.5588308672931184},{"x":42,"y":-0.4971683682878378},{"x":43,"y":-1.608182686371261},{"x":44,"y":-0.17637892642800962},{"x":45,"y":-0.9895921924637698},{"x":46,"y":-0.3046981796001678},{"x":47,"y":-0.4893167309653846},{"x":48,"y":-1.3041508158592032},{"x":49,"y":-0.41683043950842197}]}],"marks":[{"type":"line","from":{"data":"583c11d5-22f8-4df8-97dd-1d56c5e16743"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"blue"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}},{"type":"line","from":{"data":"1c8565e7-5c1f-4a3a-b581-54971972ec90"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"red"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}}]},"value":"#gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"583c11d5-22f8-4df8-97dd-1d56c5e16743\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"583c11d5-22f8-4df8-97dd-1d56c5e16743\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"583c11d5-22f8-4df8-97dd-1d56c5e16743\", :values ({:x 0, :y -0.7683} {:x 1, :y -0.37145} {:x 2, :y -1.7936} {:x 3, :y -0.067909} {:x 4, :y -0.72467} {:x 5, :y -0.7811} {:x 6, :y 0.010419} {:x 7, :y -1.5865} {:x 8, :y 0.59129} {:x 9, :y -0.99314} {:x 10, :y -1.0829} {:x 11, :y -0.17441} {:x 12, :y -0.99972} {:x 13, :y -0.8976} {:x 14, :y -0.57532} {:x 15, :y -0.92552} {:x 16, :y -0.91328} {:x 17, :y -1.1929} {:x 18, :y -0.65438} {:x 19, :y -1.657} {:x 20, :y -0.025507} {:x 21, :y -0.77629} {:x 22, :y -0.7095} {:x 23, :y 0.1119} {:x 24, :y -1.5459} {:x 25, :y 0.66598} {:x 26, :y -1.02} {:x 27, :y -0.87674} {:x 28, :y -0.15425} {:x 29, :y -1.1489} {:x 30, :y -0.82843} {:x 31, :y -0.64842} {:x 32, :y -0.82055} {:x 33, :y -1.0004} {:x 34, :y -1.0683} {:x 35, :y -0.93447} {:x 36, :y -1.2922} {:x 37, :y -0.35682} {:x 38, :y -1.3844} {:x 39, :y 0.10704} {:x 40, :y -0.93704} {:x 41, :y -0.5491} {:x 42, :y -0.474} {:x 43, :y -1.585} {:x 44, :y -0.22502} {:x 45, :y -0.97659} {:x 46, :y -0.33674} {:x 47, :y -0.60478} {:x 48, :y -1.2616} {:x 49, :y -0.52966})} {:name \"1c8565e7-5c1f-4a3a-b581-54971972ec90\", :values ({:x 0, :y -0.6561510531194803} {:x 1, :y -0.42849807387447136} {:x 2, :y -1.748285969489911} {:x 3, :y -0.02665405159870647} {:x 4, :y -0.7205401004571231} {:x 5, :y -0.7782588434464333} {:x 6, :y 0.1062867823659176} {:x 7, :y -1.5453853065686403} {:x 8, :y 0.70122101142841} {:x 9, :y -0.9862039366024559} {:x 10, :y -0.9735605431468249} {:x 11, :y -0.18177468462366195} {:x 12, :y -1.0782262963383853} {:x 13, :y -0.8794486023600043} {:x 14, :y -0.6328507740501725} {:x 15, :y -0.8539690910314582} {:x 16, :y -0.9031333403235883} {:x 17, :y -1.1723692942767856} {:x 18, :y -0.6901353453659438} {:x 19, :y -1.6055778497237165} {:x 20, :y -0.02093379587293824} {:x 21, :y -0.7630999385911313} {:x 22, :y -0.6946028370513873} {:x 23, :y 0.1298750459960506} {:x 24, :y -1.504149684160266} {:x 25, :y 0.6932713282146231} {:x 26, :y -1.0117810104295706} {:x 27, :y -0.8008590026154816} {:x 28, :y -0.15036195989404377} {:x 29, :y -1.2252812466442942} {:x 30, :y -0.8090341915281736} {:x 31, :y -0.6848263617135942} {:x 32, :y -0.7407316346822757} {:x 33, :y -0.9961792431081292} {:x 34, :y -1.0676967658287946} {:x 35, :y -0.9245987316548181} {:x 36, :y -1.2666592803119894} {:x 37, :y -0.37849975854020196} {:x 38, :y -1.356994686877228} {:x 39, :y 0.08629557617885042} {:x 40, :y -0.9094804325309619} {:x 41, :y -0.5588308672931184} {:x 42, :y -0.4971683682878378} {:x 43, :y -1.608182686371261} {:x 44, :y -0.17637892642800962} {:x 45, :y -0.9895921924637698} {:x 46, :y -0.3046981796001678} {:x 47, :y -0.4893167309653846} {:x 48, :y -1.3041508158592032} {:x 49, :y -0.41683043950842197})}), :marks ({:type \"line\", :from {:data \"583c11d5-22f8-4df8-97dd-1d56c5e16743\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"1c8565e7-5c1f-4a3a-b581-54971972ec90\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}}"},{"type":"vega","content":{"width":250,"height":154.51174926757812,"padding":{"top":10,"left":55,"bottom":40,"right":10},"scales":[{"name":"x","type":"linear","range":"width","zero":false,"domain":{"data":"53bcb6b9-7011-46d2-8857-5c0b4729270b","field":"data.x"}},{"name":"y","type":"linear","range":"height","nice":true,"zero":false,"domain":{"data":"53bcb6b9-7011-46d2-8857-5c0b4729270b","field":"data.y"}}],"axes":[{"type":"x","scale":"x","title":"Time step","titleOffset":30},{"type":"y","scale":"y"}],"data":[{"name":"53bcb6b9-7011-46d2-8857-5c0b4729270b","values":[{"x":0,"y":-0.23752},{"x":1,"y":0.98831},{"x":2,"y":0.69057},{"x":3,"y":0.88147},{"x":4,"y":-0.2074},{"x":5,"y":-0.57156},{"x":6,"y":0.84455},{"x":7,"y":0.48982},{"x":8,"y":-0.2632},{"x":9,"y":-0.76504},{"x":10,"y":-0.90355},{"x":11,"y":0.030318},{"x":12,"y":0.99796},{"x":13,"y":0.40324},{"x":14,"y":0.23665},{"x":15,"y":0.52027},{"x":16,"y":0.79671},{"x":17,"y":0.93087},{"x":18,"y":0.99681},{"x":19,"y":0.78457},{"x":20,"y":0.99837},{"x":21,"y":-0.14975},{"x":22,"y":-0.76437},{"x":23,"y":0.7783},{"x":24,"y":0.50557},{"x":25,"y":-0.42942},{"x":26,"y":-0.58685},{"x":27,"y":-0.84038},{"x":28,"y":0.24341},{"x":29,"y":0.99984},{"x":30,"y":0.48954},{"x":31,"y":0.18986},{"x":32,"y":0.48798},{"x":33,"y":0.80838},{"x":34,"y":0.96206},{"x":35,"y":0.99978},{"x":36,"y":0.97508},{"x":37,"y":0.97114},{"x":38,"y":0.33953},{"x":39,"y":0.10228},{"x":40,"y":-0.36249},{"x":41,"y":-0.30036},{"x":42,"y":0.88944},{"x":43,"y":0.62605},{"x":44,"y":0.83797},{"x":45,"y":0.012639},{"x":46,"y":-0.33613},{"x":47,"y":0.69772},{"x":48,"y":0.54889},{"x":49,"y":0.78432}]},{"name":"42b06831-6eb7-4650-811a-fb315f3d0318","values":[{"x":0,"y":-0.28495546351495227},{"x":1,"y":0.9778581236325421},{"x":2,"y":0.6615044066122386},{"x":3,"y":0.9146989374286392},{"x":4,"y":-0.20506365487501785},{"x":5,"y":-0.6526463486535596},{"x":6,"y":0.7976206326048283},{"x":7,"y":0.4864380137528287},{"x":8,"y":-0.4345278120404065},{"x":9,"y":-0.6647838175902648},{"x":10,"y":-0.8623296549439171},{"x":11,"y":0.12447257910913789},{"x":12,"y":1.0064593178680057},{"x":13,"y":0.49509589878917326},{"x":14,"y":0.26137667803940656},{"x":15,"y":0.4659031606376194},{"x":16,"y":0.7888835758407231},{"x":17,"y":0.9389304166501834},{"x":18,"y":1.0007376644662147},{"x":19,"y":0.808403444875506},{"x":20,"y":1.0049541060265847},{"x":21,"y":-0.1519157511318954},{"x":22,"y":-0.7702558428268347},{"x":23,"y":0.7629263266782141},{"x":24,"y":0.5056820991049823},{"x":25,"y":-0.495878188250454},{"x":26,"y":-0.5211053143008971},{"x":27,"y":-0.8137639973648031},{"x":28,"y":0.3038947499911813},{"x":29,"y":0.990922614541652},{"x":30,"y":0.55385849348413},{"x":31,"y":0.16525094749647143},{"x":32,"y":0.4177754320052028},{"x":33,"y":0.8214659416411162},{"x":34,"y":0.9519233198615608},{"x":35,"y":0.997401273893085},{"x":36,"y":0.982668207495889},{"x":37,"y":0.9654427232035253},{"x":38,"y":0.33439094045579926},{"x":39,"y":0.13664070037476891},{"x":40,"y":-0.40619816064734804},{"x":41,"y":-0.27169738981123676},{"x":42,"y":0.9017131280013864},{"x":43,"y":0.671198920879369},{"x":44,"y":0.9220721900112684},{"x":45,"y":9.768872998351281E-4},{"x":46,"y":-0.5058195023118203},{"x":47,"y":0.6471311485259736},{"x":48,"y":0.5509157957189816},{"x":49,"y":0.7226942050710317}]}],"marks":[{"type":"line","from":{"data":"53bcb6b9-7011-46d2-8857-5c0b4729270b"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"blue"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}},{"type":"line","from":{"data":"42b06831-6eb7-4650-811a-fb315f3d0318"},"properties":{"enter":{"x":{"scale":"x","field":"data.x"},"y":{"scale":"y","field":"data.y"},"stroke":{"value":"red"},"strokeWidth":{"value":2},"strokeOpacity":{"value":1}}}}]},"value":"#gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"53bcb6b9-7011-46d2-8857-5c0b4729270b\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"53bcb6b9-7011-46d2-8857-5c0b4729270b\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"53bcb6b9-7011-46d2-8857-5c0b4729270b\", :values ({:x 0, :y -0.23752} {:x 1, :y 0.98831} {:x 2, :y 0.69057} {:x 3, :y 0.88147} {:x 4, :y -0.2074} {:x 5, :y -0.57156} {:x 6, :y 0.84455} {:x 7, :y 0.48982} {:x 8, :y -0.2632} {:x 9, :y -0.76504} {:x 10, :y -0.90355} {:x 11, :y 0.030318} {:x 12, :y 0.99796} {:x 13, :y 0.40324} {:x 14, :y 0.23665} {:x 15, :y 0.52027} {:x 16, :y 0.79671} {:x 17, :y 0.93087} {:x 18, :y 0.99681} {:x 19, :y 0.78457} {:x 20, :y 0.99837} {:x 21, :y -0.14975} {:x 22, :y -0.76437} {:x 23, :y 0.7783} {:x 24, :y 0.50557} {:x 25, :y -0.42942} {:x 26, :y -0.58685} {:x 27, :y -0.84038} {:x 28, :y 0.24341} {:x 29, :y 0.99984} {:x 30, :y 0.48954} {:x 31, :y 0.18986} {:x 32, :y 0.48798} {:x 33, :y 0.80838} {:x 34, :y 0.96206} {:x 35, :y 0.99978} {:x 36, :y 0.97508} {:x 37, :y 0.97114} {:x 38, :y 0.33953} {:x 39, :y 0.10228} {:x 40, :y -0.36249} {:x 41, :y -0.30036} {:x 42, :y 0.88944} {:x 43, :y 0.62605} {:x 44, :y 0.83797} {:x 45, :y 0.012639} {:x 46, :y -0.33613} {:x 47, :y 0.69772} {:x 48, :y 0.54889} {:x 49, :y 0.78432})} {:name \"42b06831-6eb7-4650-811a-fb315f3d0318\", :values ({:x 0, :y -0.28495546351495227} {:x 1, :y 0.9778581236325421} {:x 2, :y 0.6615044066122386} {:x 3, :y 0.9146989374286392} {:x 4, :y -0.20506365487501785} {:x 5, :y -0.6526463486535596} {:x 6, :y 0.7976206326048283} {:x 7, :y 0.4864380137528287} {:x 8, :y -0.4345278120404065} {:x 9, :y -0.6647838175902648} {:x 10, :y -0.8623296549439171} {:x 11, :y 0.12447257910913789} {:x 12, :y 1.0064593178680057} {:x 13, :y 0.49509589878917326} {:x 14, :y 0.26137667803940656} {:x 15, :y 0.4659031606376194} {:x 16, :y 0.7888835758407231} {:x 17, :y 0.9389304166501834} {:x 18, :y 1.0007376644662147} {:x 19, :y 0.808403444875506} {:x 20, :y 1.0049541060265847} {:x 21, :y -0.1519157511318954} {:x 22, :y -0.7702558428268347} {:x 23, :y 0.7629263266782141} {:x 24, :y 0.5056820991049823} {:x 25, :y -0.495878188250454} {:x 26, :y -0.5211053143008971} {:x 27, :y -0.8137639973648031} {:x 28, :y 0.3038947499911813} {:x 29, :y 0.990922614541652} {:x 30, :y 0.55385849348413} {:x 31, :y 0.16525094749647143} {:x 32, :y 0.4177754320052028} {:x 33, :y 0.8214659416411162} {:x 34, :y 0.9519233198615608} {:x 35, :y 0.997401273893085} {:x 36, :y 0.982668207495889} {:x 37, :y 0.9654427232035253} {:x 38, :y 0.33439094045579926} {:x 39, :y 0.13664070037476891} {:x 40, :y -0.40619816064734804} {:x 41, :y -0.27169738981123676} {:x 42, :y 0.9017131280013864} {:x 43, :y 0.671198920879369} {:x 44, :y 0.9220721900112684} {:x 45, :y 9.768872998351281E-4} {:x 46, :y -0.5058195023118203} {:x 47, :y 0.6471311485259736} {:x 48, :y 0.5509157957189816} {:x 49, :y 0.7226942050710317})}), :marks ({:type \"line\", :from {:data \"53bcb6b9-7011-46d2-8857-5c0b4729270b\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"42b06831-6eb7-4650-811a-fb315f3d0318\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}}"}],"value":"[#gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"f99f98ff-dc50-4e8a-ae17-b6b8afc1bce3\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"f99f98ff-dc50-4e8a-ae17-b6b8afc1bce3\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"f99f98ff-dc50-4e8a-ae17-b6b8afc1bce3\", :values ({:x 0, :y 1.4177} {:x 1, :y 0.76228} {:x 2, :y 1.079} {:x 3, :y -0.20891} {:x 4, :y -0.6084} {:x 5, :y 1.0057} {:x 6, :y 0.51189} {:x 7, :y -0.26634} {:x 8, :y -0.8711} {:x 9, :y -1.128} {:x 10, :y 0.030323} {:x 11, :y 1.5069} {:x 12, :y 0.41506} {:x 13, :y 0.23892} {:x 14, :y 0.54717} {:x 15, :y 0.92183} {:x 16, :y 1.1968} {:x 17, :y 1.6507} {:x 18, :y 0.90201} {:x 19, :y 1.6279} {:x 20, :y -0.15032} {:x 21, :y -0.87007} {:x 22, :y 0.89195} {:x 23, :y 0.53004} {:x 24, :y -0.44385} {:x 25, :y -0.62716} {:x 26, :y -0.99799} {:x 27, :y 0.24588} {:x 28, :y 1.5889} {:x 29, :y 0.51156} {:x 30, :y 0.19102} {:x 31, :y 0.50977} {:x 32, :y 0.9414} {:x 33, :y 1.2945} {:x 34, :y 1.5499} {:x 35, :y 1.3471} {:x 36, :y 1.8116} {:x 37, :y 0.34641} {:x 38, :y 0.10246} {:x 39, :y -0.37094} {:x 40, :y -0.30507} {:x 41, :y 1.0961} {:x 42, :y 0.67648} {:x 43, :y 0.99356} {:x 44, :y 0.012639} {:x 45, :y -0.3428} {:x 46, :y 0.77221} {:x 47, :y 0.58104} {:x 48, :y 0.9016} {:x 49, :y 0.58413})} {:name \"5b09fd8d-b39d-44c8-9140-4d91ad8318c1\", :values ({:x 0, :y 1.47789421139643} {:x 1, :y 0.7485151024684631} {:x 2, :y 1.1256099629495988} {:x 3, :y -0.21853907378491716} {:x 4, :y -0.7269873612328437} {:x 5, :y 0.9243906448806127} {:x 6, :y 0.5079399323915975} {:x 7, :y -0.4643609909305077} {:x 8, :y -0.6938857135381984} {:x 9, :y -1.0751172609556205} {:x 10, :y 0.11127752883728341} {:x 11, :y 1.5650434986725124} {:x 12, :y 0.5098879052689004} {:x 13, :y 0.2700731527855736} {:x 14, :y 0.4865271103826882} {:x 15, :y 0.891685387701729} {:x 16, :y 1.183537517436241} {:x 17, :y 1.6181818165324042} {:x 18, :y 0.9212837951635541} {:x 19, :y 1.6425641829230153} {:x 20, :y -0.1637719679958235} {:x 21, :y -0.8777910035632345} {:x 22, :y 0.886981748947608} {:x 23, :y 0.5247962932184906} {:x 24, :y -0.508562527666783} {:x 25, :y -0.5699526124191071} {:x 26, :y -0.9362095719068727} {:x 27, :y 0.31227804696536793} {:x 28, :y 1.501237006245938} {:x 29, :y 0.6093597664377434} {:x 30, :y 0.18565776324740893} {:x 31, :y 0.44712944601659305} {:x 32, :y 0.9400324458839735} {:x 33, :y 1.2674898358942575} {:x 34, :y 1.5182821258431827} {:x 35, :y 1.3356908825967275} {:x 36, :y 1.7984611996374504} {:x 37, :y 0.3372543931653316} {:x 38, :y 0.14931963354971414} {:x 39, :y -0.4049218506098873} {:x 40, :y -0.266311495138262} {:x 41, :y 1.137697659887803} {:x 42, :y 0.727910766774785} {:x 43, :y 1.1380352685066726} {:x 44, :y 0.014122045134247082} {:x 45, :y -0.5196234742375875} {:x 46, :y 0.6973920033661354} {:x 47, :y 0.576054622226453} {:x 48, :y 0.8254914141780186} {:x 49, :y 0.3288001151069147})}), :marks ({:type \"line\", :from {:data \"f99f98ff-dc50-4e8a-ae17-b6b8afc1bce3\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"5b09fd8d-b39d-44c8-9140-4d91ad8318c1\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}} #gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"583c11d5-22f8-4df8-97dd-1d56c5e16743\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"583c11d5-22f8-4df8-97dd-1d56c5e16743\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"583c11d5-22f8-4df8-97dd-1d56c5e16743\", :values ({:x 0, :y -0.7683} {:x 1, :y -0.37145} {:x 2, :y -1.7936} {:x 3, :y -0.067909} {:x 4, :y -0.72467} {:x 5, :y -0.7811} {:x 6, :y 0.010419} {:x 7, :y -1.5865} {:x 8, :y 0.59129} {:x 9, :y -0.99314} {:x 10, :y -1.0829} {:x 11, :y -0.17441} {:x 12, :y -0.99972} {:x 13, :y -0.8976} {:x 14, :y -0.57532} {:x 15, :y -0.92552} {:x 16, :y -0.91328} {:x 17, :y -1.1929} {:x 18, :y -0.65438} {:x 19, :y -1.657} {:x 20, :y -0.025507} {:x 21, :y -0.77629} {:x 22, :y -0.7095} {:x 23, :y 0.1119} {:x 24, :y -1.5459} {:x 25, :y 0.66598} {:x 26, :y -1.02} {:x 27, :y -0.87674} {:x 28, :y -0.15425} {:x 29, :y -1.1489} {:x 30, :y -0.82843} {:x 31, :y -0.64842} {:x 32, :y -0.82055} {:x 33, :y -1.0004} {:x 34, :y -1.0683} {:x 35, :y -0.93447} {:x 36, :y -1.2922} {:x 37, :y -0.35682} {:x 38, :y -1.3844} {:x 39, :y 0.10704} {:x 40, :y -0.93704} {:x 41, :y -0.5491} {:x 42, :y -0.474} {:x 43, :y -1.585} {:x 44, :y -0.22502} {:x 45, :y -0.97659} {:x 46, :y -0.33674} {:x 47, :y -0.60478} {:x 48, :y -1.2616} {:x 49, :y -0.52966})} {:name \"1c8565e7-5c1f-4a3a-b581-54971972ec90\", :values ({:x 0, :y -0.6561510531194803} {:x 1, :y -0.42849807387447136} {:x 2, :y -1.748285969489911} {:x 3, :y -0.02665405159870647} {:x 4, :y -0.7205401004571231} {:x 5, :y -0.7782588434464333} {:x 6, :y 0.1062867823659176} {:x 7, :y -1.5453853065686403} {:x 8, :y 0.70122101142841} {:x 9, :y -0.9862039366024559} {:x 10, :y -0.9735605431468249} {:x 11, :y -0.18177468462366195} {:x 12, :y -1.0782262963383853} {:x 13, :y -0.8794486023600043} {:x 14, :y -0.6328507740501725} {:x 15, :y -0.8539690910314582} {:x 16, :y -0.9031333403235883} {:x 17, :y -1.1723692942767856} {:x 18, :y -0.6901353453659438} {:x 19, :y -1.6055778497237165} {:x 20, :y -0.02093379587293824} {:x 21, :y -0.7630999385911313} {:x 22, :y -0.6946028370513873} {:x 23, :y 0.1298750459960506} {:x 24, :y -1.504149684160266} {:x 25, :y 0.6932713282146231} {:x 26, :y -1.0117810104295706} {:x 27, :y -0.8008590026154816} {:x 28, :y -0.15036195989404377} {:x 29, :y -1.2252812466442942} {:x 30, :y -0.8090341915281736} {:x 31, :y -0.6848263617135942} {:x 32, :y -0.7407316346822757} {:x 33, :y -0.9961792431081292} {:x 34, :y -1.0676967658287946} {:x 35, :y -0.9245987316548181} {:x 36, :y -1.2666592803119894} {:x 37, :y -0.37849975854020196} {:x 38, :y -1.356994686877228} {:x 39, :y 0.08629557617885042} {:x 40, :y -0.9094804325309619} {:x 41, :y -0.5588308672931184} {:x 42, :y -0.4971683682878378} {:x 43, :y -1.608182686371261} {:x 44, :y -0.17637892642800962} {:x 45, :y -0.9895921924637698} {:x 46, :y -0.3046981796001678} {:x 47, :y -0.4893167309653846} {:x 48, :y -1.3041508158592032} {:x 49, :y -0.41683043950842197})}), :marks ({:type \"line\", :from {:data \"583c11d5-22f8-4df8-97dd-1d56c5e16743\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"1c8565e7-5c1f-4a3a-b581-54971972ec90\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}} #gorilla_repl.vega.VegaView{:content {:width 250, :height 154.51175, :padding {:top 10, :left 55, :bottom 40, :right 10}, :scales [{:name \"x\", :type \"linear\", :range \"width\", :zero false, :domain {:data \"53bcb6b9-7011-46d2-8857-5c0b4729270b\", :field \"data.x\"}} {:name \"y\", :type \"linear\", :range \"height\", :nice true, :zero false, :domain {:data \"53bcb6b9-7011-46d2-8857-5c0b4729270b\", :field \"data.y\"}}], :axes [{:type \"x\", :scale \"x\", :title \"Time step\", :titleOffset 30} {:type \"y\", :scale \"y\"}], :data ({:name \"53bcb6b9-7011-46d2-8857-5c0b4729270b\", :values ({:x 0, :y -0.23752} {:x 1, :y 0.98831} {:x 2, :y 0.69057} {:x 3, :y 0.88147} {:x 4, :y -0.2074} {:x 5, :y -0.57156} {:x 6, :y 0.84455} {:x 7, :y 0.48982} {:x 8, :y -0.2632} {:x 9, :y -0.76504} {:x 10, :y -0.90355} {:x 11, :y 0.030318} {:x 12, :y 0.99796} {:x 13, :y 0.40324} {:x 14, :y 0.23665} {:x 15, :y 0.52027} {:x 16, :y 0.79671} {:x 17, :y 0.93087} {:x 18, :y 0.99681} {:x 19, :y 0.78457} {:x 20, :y 0.99837} {:x 21, :y -0.14975} {:x 22, :y -0.76437} {:x 23, :y 0.7783} {:x 24, :y 0.50557} {:x 25, :y -0.42942} {:x 26, :y -0.58685} {:x 27, :y -0.84038} {:x 28, :y 0.24341} {:x 29, :y 0.99984} {:x 30, :y 0.48954} {:x 31, :y 0.18986} {:x 32, :y 0.48798} {:x 33, :y 0.80838} {:x 34, :y 0.96206} {:x 35, :y 0.99978} {:x 36, :y 0.97508} {:x 37, :y 0.97114} {:x 38, :y 0.33953} {:x 39, :y 0.10228} {:x 40, :y -0.36249} {:x 41, :y -0.30036} {:x 42, :y 0.88944} {:x 43, :y 0.62605} {:x 44, :y 0.83797} {:x 45, :y 0.012639} {:x 46, :y -0.33613} {:x 47, :y 0.69772} {:x 48, :y 0.54889} {:x 49, :y 0.78432})} {:name \"42b06831-6eb7-4650-811a-fb315f3d0318\", :values ({:x 0, :y -0.28495546351495227} {:x 1, :y 0.9778581236325421} {:x 2, :y 0.6615044066122386} {:x 3, :y 0.9146989374286392} {:x 4, :y -0.20506365487501785} {:x 5, :y -0.6526463486535596} {:x 6, :y 0.7976206326048283} {:x 7, :y 0.4864380137528287} {:x 8, :y -0.4345278120404065} {:x 9, :y -0.6647838175902648} {:x 10, :y -0.8623296549439171} {:x 11, :y 0.12447257910913789} {:x 12, :y 1.0064593178680057} {:x 13, :y 0.49509589878917326} {:x 14, :y 0.26137667803940656} {:x 15, :y 0.4659031606376194} {:x 16, :y 0.7888835758407231} {:x 17, :y 0.9389304166501834} {:x 18, :y 1.0007376644662147} {:x 19, :y 0.808403444875506} {:x 20, :y 1.0049541060265847} {:x 21, :y -0.1519157511318954} {:x 22, :y -0.7702558428268347} {:x 23, :y 0.7629263266782141} {:x 24, :y 0.5056820991049823} {:x 25, :y -0.495878188250454} {:x 26, :y -0.5211053143008971} {:x 27, :y -0.8137639973648031} {:x 28, :y 0.3038947499911813} {:x 29, :y 0.990922614541652} {:x 30, :y 0.55385849348413} {:x 31, :y 0.16525094749647143} {:x 32, :y 0.4177754320052028} {:x 33, :y 0.8214659416411162} {:x 34, :y 0.9519233198615608} {:x 35, :y 0.997401273893085} {:x 36, :y 0.982668207495889} {:x 37, :y 0.9654427232035253} {:x 38, :y 0.33439094045579926} {:x 39, :y 0.13664070037476891} {:x 40, :y -0.40619816064734804} {:x 41, :y -0.27169738981123676} {:x 42, :y 0.9017131280013864} {:x 43, :y 0.671198920879369} {:x 44, :y 0.9220721900112684} {:x 45, :y 9.768872998351281E-4} {:x 46, :y -0.5058195023118203} {:x 47, :y 0.6471311485259736} {:x 48, :y 0.5509157957189816} {:x 49, :y 0.7226942050710317})}), :marks ({:type \"line\", :from {:data \"53bcb6b9-7011-46d2-8857-5c0b4729270b\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"blue\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}} {:type \"line\", :from {:data \"42b06831-6eb7-4650-811a-fb315f3d0318\"}, :properties {:enter {:x {:scale \"x\", :field \"data.x\"}, :y {:scale \"y\", :field \"data.y\"}, :stroke {:value \"red\"}, :strokeWidth {:value 2}, :strokeOpacity {:value 1}}}})}}]"}
;; <=

;; @@

;; @@
